# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pyright: reportPrivateUsage=none
# pyright: reportUnknownMemberType=none
# This leads to too many false-positives around mocks.

import asyncio
import datetime
import functools
import itertools
import time
import traceback
import types
import typing
from collections import abc as collections
from unittest import mock

import alluka
import freezegun
import pytest

import tanjun

_CallbackSig = collections.Callable[..., collections.Coroutine[typing.Any, typing.Any, typing.Any]]
_CallbackSigT = typing.TypeVar("_CallbackSigT", bound=_CallbackSig)
_T = typing.TypeVar("_T")
_TIMEOUT: typing.Final[float] = 10.0


def _chain(data: collections.Iterable[collections.Iterable[_T]]) -> list[_T]:
    return list(itertools.chain.from_iterable(data))


def _print_tb(callback: _CallbackSigT, /) -> _CallbackSigT:
    @functools.wraps(callback)
    async def wrapper(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        try:
            return await callback(*args, **kwargs)
        except Exception:
            traceback.print_exc()
            raise

    return typing.cast("_CallbackSigT", wrapper)


class _ManualClock:
    def __init__(
        self,
        freeze_time: freezegun.api.FrozenDateTimeFactory,
        tick_fors: list[datetime.timedelta],
        *,
        interval_ratio: int = 10,
        post_sleep_count: int = 5,
        tick_sleep_count: int = 1,
    ) -> None:
        self._freeze_time = freeze_time
        self._index = -1
        self._interval_ratio = interval_ratio
        self._is_ticking = False
        self._keep_ticking = False
        self._post_sleep_count = post_sleep_count
        self._tasks: list[asyncio.Task[None]] = []
        self._tick_fors = tick_fors
        self._tick_sleep_count = tick_sleep_count

    def spawn_ticker(self) -> "_ManualClock":
        if self._is_ticking and self._keep_ticking:
            raise RuntimeError("Already ticking")

        if self._is_ticking:
            self._keep_ticking = True
            return self

        self._is_ticking = True
        self._tasks.append(asyncio.get_running_loop().create_task(_print_tb(self._next_tick)()))
        return self

    def stop_ticker(self) -> None:
        self._tasks[-1].cancel()

    async def _next_tick(self) -> None:
        index = self._index + 1
        if len(self._tick_fors) == index:
            tick_for = self._tick_fors[-1]
        else:
            tick_for = self._tick_fors[index]
            self._index += 1

        interval = tick_for / self._interval_ratio
        while tick_for > datetime.timedelta():
            for _ in range(self._tick_sleep_count):
                # This lets the event loop run for a bit between ticks.
                await asyncio.sleep(0)

            if tick_for - interval >= datetime.timedelta():
                self._freeze_time.tick(interval)
            else:
                self._freeze_time.tick(tick_for)

            tick_for -= interval

        for _ in range(self._post_sleep_count):
            # This lets the event loop run for a bit between ticks.
            await asyncio.sleep(0)

        if self._keep_ticking:
            self._keep_ticking = False
            return await self._next_tick()

        self._is_ticking = False


def test_as_interval():
    mock_callback = mock.Mock()

    result = tanjun.as_interval(123, fatal_exceptions=[KeyError], ignored_exceptions=[TabError], max_runs=55)(
        mock_callback
    )

    assert result.interval == datetime.timedelta(minutes=2, seconds=3)
    assert result._fatal_exceptions == (KeyError,)
    assert result._ignored_exceptions == (TabError,)
    assert result._max_runs == 55
    assert isinstance(result, tanjun.schedules.IntervalSchedule)


class TestIntervalSchedule:
    def test_callback_property(self):
        mock_callback = mock.Mock()
        interval = tanjun.schedules.IntervalSchedule(mock_callback, 123)

        assert interval.callback is mock_callback

    def test_is_alive(self):
        assert tanjun.schedules.IntervalSchedule(mock.Mock(), 34123).is_alive is False

    def test_is_alive_when_is_alive(self):
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123)
        interval._task = mock.Mock()

        assert interval.is_alive is True

    @pytest.mark.parametrize(
        "interval", [datetime.timedelta(days=7, hours=13, minutes=8, seconds=54), 652134, 652134.0]
    )
    def test_interval_property(self, interval: typing.Union[int, float, datetime.timedelta]):
        interval_ = tanjun.schedules.IntervalSchedule(mock.Mock(), interval)

        assert interval_.interval == datetime.timedelta(days=7, hours=13, minutes=8, seconds=54)

    def test_iteration_count_property(self):
        assert tanjun.schedules.IntervalSchedule(mock.Mock(), 123).iteration_count == 0

    @pytest.mark.asyncio()
    async def test_call_dunder_method(self):
        mock_callback = mock.AsyncMock()
        interval = tanjun.schedules.IntervalSchedule(typing.cast("tanjun.schedules._CallbackSig", mock_callback), 123)

        await interval(123, 543, sex="OK", boo="31123")

        mock_callback.assert_awaited_once_with(123, 543, sex="OK", boo="31123")

    def test_copy(self):
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123)
        interval._tasks.append(mock.Mock())

        result = interval.copy()

        assert result.callback is interval.callback
        assert result.interval is interval.interval
        assert result.iteration_count is interval.iteration_count
        assert result._tasks == []
        assert result._tasks is not interval._tasks
        assert result is not interval

    def test_copy_when_schedule_is_active(self):
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123)
        interval._task = mock.Mock()

        with pytest.raises(RuntimeError, match="Cannot copy an active schedule"):
            interval.copy()

    def test_load_into_component(self):
        mock_component = mock.Mock(tanjun.Component)
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123)

        interval.load_into_component(mock_component)

        mock_component.add_schedule.assert_called_once_with(interval)

    def test_load_into_component_when_no_add_schedule_method(self):
        mock_component = mock.Mock(object)
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123)

        interval.load_into_component(mock_component)

    def test_set_start_callback(self):
        mock_callback = mock.Mock()
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123)

        result = interval.set_start_callback(mock_callback)

        assert result is interval
        assert interval._start_callback
        assert interval._start_callback is mock_callback

    def test_set_stop_callback(self):
        mock_callback = mock.Mock()
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123)

        result = interval.set_stop_callback(mock_callback)

        assert result is interval
        assert interval._stop_callback
        assert interval._stop_callback is mock_callback

    @pytest.mark.asyncio()
    async def test__execute(self):
        mock_client = mock.AsyncMock()
        mock_callback = mock.Mock()
        stop = mock.Mock()
        interval: tanjun.schedules.IntervalSchedule[typing.Any] = types.new_class(
            "StubIntervalSchedule",
            (tanjun.schedules.IntervalSchedule[typing.Any],),
            exec_body=lambda ns: ns.update({"stop": stop}),
        )(mock_callback, 123)

        await interval._execute(mock_client)

        mock_client.call_with_async_di.assert_awaited_once_with(mock_callback)
        stop.assert_not_called()

    @pytest.mark.asyncio()
    async def test__execute_when_fatal_exception(self):
        mock_callback = mock.Mock()
        mock_client = mock.AsyncMock()
        mock_client.call_with_async_di.side_effect = KeyError("hihihiih")
        stop = mock.AsyncMock()
        interval: tanjun.schedules.IntervalSchedule[typing.Any] = types.new_class(
            "StubIntervalSchedule",
            (tanjun.schedules.IntervalSchedule[typing.Any],),
            exec_body=lambda ns: ns.update({"stop": stop}),
        )(mock_callback, 123, fatal_exceptions=[LookupError], ignored_exceptions=[Exception])

        await interval._execute(mock_client)

        mock_client.call_with_async_di.assert_awaited_once_with(mock_callback)
        stop.assert_awaited_once_with()
        assert interval._tasks == []

    @pytest.mark.asyncio()
    async def test__execute_when_ignored_exception(self):
        mock_callback = mock.Mock()
        mock_client = mock.AsyncMock()
        mock_client.call_with_async_di.side_effect = IndexError("hihihiih")
        stop = mock.Mock()
        interval: tanjun.schedules.IntervalSchedule[typing.Any] = types.new_class(
            "StubIntervalSchedule",
            (tanjun.schedules.IntervalSchedule[typing.Any],),
            exec_body=lambda ns: ns.update({"stop": stop}),
        )(mock_callback, 123, fatal_exceptions=[KeyError], ignored_exceptions=[LookupError])

        await interval._execute(mock_client)

        mock_client.call_with_async_di.assert_awaited_once_with(mock_callback)
        stop.assert_not_called()

    @pytest.mark.asyncio()
    async def test__execute_when_exception(self):
        mock_callback = mock.Mock()
        mock_client = mock.AsyncMock()
        error = ValueError("hihihiih")
        mock_client.call_with_async_di.side_effect = error
        stop = mock.Mock()
        interval: tanjun.schedules.IntervalSchedule[typing.Any] = types.new_class(
            "StubIntervalSchedule",
            (tanjun.schedules.IntervalSchedule[typing.Any],),
            exec_body=lambda ns: ns.update({"stop": stop}),
        )(mock_callback, 123, fatal_exceptions=[KeyError], ignored_exceptions=[TypeError])

        await interval._execute(mock_client)

        mock_client.call_with_async_di.assert_awaited_once_with(mock_callback)
        stop.assert_not_called()

    @pytest.mark.timeout(_TIMEOUT)
    @pytest.mark.asyncio()
    async def test__loop(self):
        mock_client = alluka.Client()
        call_times: list[int] = []

        @_print_tb
        async def callback():
            call_times.append(time.time_ns())
            clock.spawn_ticker()

        interval = tanjun.schedules.IntervalSchedule(callback, 5, ignored_exceptions=[LookupError])

        with freezegun.freeze_time(datetime.datetime(2012, 1, 14, 12)) as frozen_time:
            assert isinstance(frozen_time, freezegun.api.FrozenDateTimeFactory)
            interval.start(mock_client)
            # Note: these have to be at least a microsecond after the target time as
            # the unix event loop won't return the sleep until the target time has passed,
            # not just been reached.
            clock = _ManualClock(frozen_time, [datetime.timedelta(seconds=5, milliseconds=100)]).spawn_ticker()
            await asyncio.sleep(30)
            await interval.stop()
            clock.stop_ticker()

        assert interval._task is None
        assert call_times == [
            1326542405000000000,
            1326542410000000000,
            1326542415000000000,
            1326542420000000000,
            1326542425000000000,
        ]
        assert interval.iteration_count == 5

    @pytest.mark.timeout(_TIMEOUT)
    @pytest.mark.asyncio()
    async def test__loop_when_max_runs(self):
        mock_client = alluka.Client()
        call_times: list[int] = []
        close_event = asyncio.Event()
        close_time: typing.Optional[int] = None

        @_print_tb
        async def callback():
            call_times.append(time.time_ns())
            clock.spawn_ticker()

        @_print_tb
        async def on_stop():
            nonlocal close_time
            close_time = time.time_ns()
            close_event.set()

        interval = tanjun.schedules.IntervalSchedule(
            callback, 3, ignored_exceptions=[LookupError], max_runs=3
        ).set_stop_callback(on_stop)

        with freezegun.freeze_time(datetime.datetime(2012, 4, 11, 12)) as frozen_time:
            assert isinstance(frozen_time, freezegun.api.FrozenDateTimeFactory)
            interval.start(mock_client)
            # Note: these have to be at least a microsecond after the target time as
            # the unix event loop won't return the sleep until the target time has passed,
            # not just been reached.
            clock = _ManualClock(frozen_time, [datetime.timedelta(seconds=3, milliseconds=300)]).spawn_ticker()
            await close_event.wait()
            assert close_time == 1334145609000000000

        # This is done to allow any finished tasks to be removed.
        await asyncio.sleep(0)

        assert interval._task is None
        assert interval._tasks == []
        assert call_times == [1334145603000000000, 1334145606000000000, 1334145609000000000]
        assert interval.iteration_count == 3

    @pytest.mark.timeout(_TIMEOUT)
    @pytest.mark.asyncio()
    async def test__loop_when_start_and_stop_callbacks_set(self):
        mock_client = alluka.Client()
        call_times: list[int] = []
        start_time = None
        stop_time = None

        @_print_tb
        async def callback() -> None:
            call_times.append(time.time_ns())
            clock.spawn_ticker()

        @_print_tb
        async def on_start() -> None:
            nonlocal start_time
            start_time = time.time_ns()

        @_print_tb
        async def on_stop() -> None:
            nonlocal stop_time
            stop_time = time.time_ns()
            clock.stop_ticker()

        interval = (
            tanjun.schedules.IntervalSchedule(callback, 7, ignored_exceptions=[LookupError])
            .set_start_callback(on_start)
            .set_stop_callback(on_stop)
        )

        with freezegun.freeze_time(datetime.datetime(2011, 4, 5, 4)) as frozen_time:
            assert isinstance(frozen_time, freezegun.api.FrozenDateTimeFactory)
            interval.start(mock_client)
            # Note: these have to be at least a microsecond after the target time as
            # the unix event loop won't return the sleep until the target time has passed,
            # not just been reached.
            clock = _ManualClock(frozen_time, [datetime.timedelta(seconds=7, milliseconds=200)]).spawn_ticker()
            await asyncio.sleep(28)
            await interval.stop()
            await asyncio.sleep(0)

        assert interval._task is None
        assert call_times == [1301976007000000000, 1301976014000000000, 1301976021000000000]
        assert start_time == 1301976000000000000
        assert stop_time == 1301976028000000000
        assert interval.iteration_count == 3

    @pytest.mark.parametrize("fatal_exceptions", [[LookupError], []])
    @pytest.mark.asyncio()
    async def test__loop_and_start_raises(self, fatal_exceptions: list[type[Exception]]):
        error = KeyError()
        mock_client = mock.Mock()
        mock_client.call_with_async_di = mock.AsyncMock(side_effect=error)
        mock_client.get_callback_override.return_value = None
        mock_start = mock.Mock()
        mock_execute = mock.Mock()
        interval: tanjun.schedules.IntervalSchedule[typing.Any] = types.new_class(
            "StubIntervalSchedule",
            (tanjun.schedules.IntervalSchedule[typing.Any],),
            exec_body=lambda ns: ns.update({"_execute": mock_execute}),
        )(mock.Mock(), 123, ignored_exceptions=[RuntimeError], fatal_exceptions=fatal_exceptions).set_start_callback(
            mock_start
        )
        interval._task = mock.Mock()

        with mock.patch.object(asyncio, "sleep") as sleep:
            await interval._loop(mock_client)

        mock_client.call_with_async_di.assert_awaited_once_with(mock_start)
        mock_execute.assert_not_called()
        sleep.assert_not_called()
        assert interval._task is None

    @pytest.mark.asyncio()
    async def test__loop_and_start_raises_ignored(self):
        mock_client = mock.Mock()
        mock_client.call_with_async_di = mock.AsyncMock(side_effect=KeyError())
        mock_client.get_callback_override.return_value = None
        mock_start = mock.Mock()
        mock_execute = mock.Mock()
        interval: tanjun.schedules.IntervalSchedule[typing.Any] = types.new_class(
            "StubIntervalSchedule",
            (tanjun.schedules.IntervalSchedule[typing.Any],),
            exec_body=lambda ns: ns.update({"_execute": mock_execute}),
        )(mock.Mock(), 123, ignored_exceptions=[LookupError]).set_start_callback(mock_start)
        interval._task = mock.Mock()

        with (
            mock.patch.object(asyncio, "sleep", side_effect=asyncio.CancelledError) as sleep,
            pytest.raises(asyncio.CancelledError),
            mock.patch.object(asyncio, "get_running_loop") as get_running_loop,
        ):
            await interval._loop(mock_client)

        mock_client.call_with_async_di.assert_awaited_once_with(mock_start)
        mock_execute.assert_not_called()
        get_running_loop.return_value.create_task.assert_not_called()
        sleep.assert_called_once_with(123.0)

    def test_start(self):
        mock_client = mock.Mock()
        loop_method = mock.Mock()
        interval: tanjun.schedules.IntervalSchedule[typing.Any] = types.new_class(
            "StubIntervalSchedule",
            (tanjun.schedules.IntervalSchedule[typing.Any],),
            exec_body=lambda ns: ns.update({"_loop": loop_method}),
        )(mock.Mock(), 123)

        with mock.patch.object(asyncio, "get_running_loop") as get_running_loop:
            interval.start(mock_client)

        assert interval._task is get_running_loop.return_value.create_task.return_value
        assert interval._client is mock_client
        assert interval.is_alive is True
        get_running_loop.return_value.create_task.assert_called_once_with(loop_method.return_value)
        get_running_loop.assert_called_once_with()

    def test_start_when_passed_event_loop(self):
        mock_client = mock.Mock()
        mock_loop = mock.Mock()
        loop_method = mock.Mock()
        interval: tanjun.schedules.IntervalSchedule[typing.Any] = types.new_class(
            "StubIntervalSchedule",
            (tanjun.schedules.IntervalSchedule[typing.Any],),
            exec_body=lambda ns: ns.update({"_loop": loop_method}),
        )(mock.Mock(), 123)

        interval.start(mock_client, loop=mock_loop)

        assert interval._task is mock_loop.create_task.return_value
        assert interval._client is mock_client
        assert interval.is_alive is True
        mock_loop.create_task.assert_called_once_with(loop_method.return_value)

    def test_start_when_passed_event_loop_isnt_active(self):
        mock_loop = mock.Mock()
        mock_loop.is_running.return_value = False
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123)

        with pytest.raises(RuntimeError, match="Event loop is not running"):
            interval.start(mock.Mock(), loop=mock_loop)

        assert interval._task is None

    def test_start_when_already_active(self):
        mock_task = mock.Mock()
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123)
        interval._task = mock_task

        with pytest.raises(RuntimeError, match="Cannot start an active schedule"):
            interval.start(mock.Mock())

        assert interval._task is mock_task

    @pytest.mark.asyncio()
    async def test_force_stop(self):
        mock_task = mock.Mock()
        mock_task_1 = asyncio.create_task(asyncio.sleep(2, result=None))
        mock_task_2 = asyncio.create_task(asyncio.sleep(2, result=None))
        mock_task_3 = asyncio.create_task(asyncio.sleep(2, result=None))
        mock_client = mock.AsyncMock()
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123)
        interval._client = mock_client
        interval._task = mock_task
        interval._add_task(mock_task_1)
        interval._add_task(mock_task_2)
        interval._add_task(mock_task_3)

        interval.force_stop()
        # This is done to allow any finished tasks to be removed.
        await asyncio.sleep(0.1)

        mock_task.cancel.assert_called_once_with()
        assert interval.is_alive is False
        assert interval._task is None
        assert interval._tasks == []
        assert mock_task_1.cancelled() is True
        assert mock_task_2.cancelled() is True
        assert mock_task_3.cancelled() is True
        mock_client.call_with_async_di.assert_not_called()

    @pytest.mark.asyncio()
    async def test_force_stop_when_stop_callback_set(self):
        mock_task = mock.Mock()
        mock_client = mock.AsyncMock()
        mock_stop_callback = mock.Mock()
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123).set_stop_callback(mock_stop_callback)
        interval._client = mock_client
        interval._task = mock_task
        interval._tasks = []

        interval.force_stop()
        # Await the stop callback task.
        await interval._tasks[-1]

        mock_task.cancel.assert_called_once_with()
        assert interval._task is None
        assert interval._tasks == []
        mock_client.call_with_async_di.assert_awaited_once_with(mock_stop_callback)

    @pytest.mark.asyncio()
    async def test_force_stop_when_stop_callback_stop_raises(self):
        mock_task = mock.Mock()
        mock_client = mock.AsyncMock()
        mock_client.call_with_async_di.side_effect = KeyError
        mock_stop_callback = mock.Mock()
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123).set_stop_callback(mock_stop_callback)
        interval._client = mock_client
        interval._task = mock_task
        interval._add_task(asyncio.create_task(asyncio.sleep(50, result=None)))

        interval.force_stop()
        # Await the stop callback task.
        await interval._tasks[-1]

        mock_task.cancel.assert_called_once_with()
        assert interval._task is None
        assert interval._tasks == []
        mock_client.call_with_async_di.assert_awaited_once_with(mock_stop_callback)

    @pytest.mark.asyncio()
    async def test_force_stop_when_stop_callback_raises_ignored(self):
        mock_task = mock.Mock()
        mock_client = mock.AsyncMock()
        mock_client.call_with_async_di.side_effect = ValueError
        mock_stop_callback = mock.Mock()
        interval = tanjun.schedules.IntervalSchedule(
            mock.Mock(), 123, ignored_exceptions=(ValueError,)
        ).set_stop_callback(mock_stop_callback)
        interval._client = mock_client
        interval._task = mock_task
        interval._tasks = []

        interval.force_stop()

        mock_task.cancel.assert_called_once_with()
        assert interval._task is None
        assert len(interval._tasks) == 1
        await interval._tasks[0]
        assert interval._tasks == []
        mock_client.call_with_async_di.assert_awaited_once_with(mock_stop_callback)

    @pytest.mark.asyncio()
    async def test_force_stop_when_no_tasks(self):
        mock_task = mock.Mock()
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123)
        interval._client = mock.AsyncMock()
        interval._task = mock_task
        interval._tasks = []

        interval.force_stop()

        mock_task.cancel.assert_called_once_with()
        assert interval._task is None
        assert interval._tasks == []

    def test_force_stop_when_not_active(self):
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123)

        with pytest.raises(RuntimeError, match="Schedule is not running"):
            interval.force_stop()

    @pytest.mark.asyncio()
    async def test_stop(self):
        mock_task = mock.Mock()
        mock_client = mock.AsyncMock()
        mock_task_1 = asyncio.create_task(asyncio.sleep(0.2, result=None))
        mock_task_2 = asyncio.create_task(asyncio.sleep(0.2, result=None))
        mock_task_3 = asyncio.create_task(asyncio.sleep(0.2, result=None))
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123)
        interval._client = mock_client
        interval._task = mock_task
        interval._add_task(mock_task_1)
        interval._add_task(mock_task_2)
        interval._add_task(mock_task_3)

        await interval.stop()
        # This is done to allow any finished tasks to be removed.
        await asyncio.sleep(0)

        mock_task.cancel.assert_called_once_with()
        assert interval.is_alive is False
        assert interval._task is None
        assert interval._tasks == []
        assert mock_task_1.result() is None
        assert mock_task_2.result() is None
        assert mock_task_3.result() is None
        mock_client.call_with_async_di.assert_not_called()

    @pytest.mark.asyncio()
    async def test_stop_when_no_tasks(self):
        mock_task = mock.Mock()
        mock_client = mock.AsyncMock()
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123)
        interval._client = mock_client
        interval._task = mock_task
        interval._tasks = []

        await interval.stop()
        # This is done to allow any finished tasks to be removed.
        await asyncio.sleep(0)

        mock_task.cancel.assert_called_once_with()
        assert interval._task is None
        assert interval._tasks == []
        mock_client.call_with_async_di.assert_not_called()

    @pytest.mark.asyncio()
    async def test_stop_when_some_tasks_time_out(self):
        mock_task = mock.Mock()
        mock_client = mock.AsyncMock()
        mock_task_1 = asyncio.create_task(asyncio.sleep(0.1, result=None))
        mock_task_2 = asyncio.create_task(asyncio.sleep(0.7, result=None))
        mock_task_3 = asyncio.create_task(asyncio.sleep(0.2, result=None))
        mock_task_4 = asyncio.create_task(asyncio.sleep(0.8, result=None))
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123)
        interval._client = mock_client
        interval._task = mock_task
        interval._add_task(mock_task_1)
        interval._add_task(mock_task_2)
        interval._add_task(mock_task_3)
        interval._add_task(mock_task_4)

        with pytest.raises(asyncio.TimeoutError):
            await asyncio.wait_for(interval.stop(), 0.3)

        # This is done to allow any finished tasks to be removed.
        await asyncio.sleep(0)

        mock_task.cancel.assert_called_once_with()
        assert interval.is_alive is False
        assert interval._task is None
        assert interval._tasks == []
        assert mock_task_1.result() is None
        assert mock_task_2.cancelled() is True
        assert mock_task_3.result() is None
        assert mock_task_4.cancelled() is True
        mock_client.call_with_async_di.assert_not_called()

    @pytest.mark.asyncio()
    async def test_stop_when_stop_callback_stop_set(self):
        mock_task = mock.Mock()
        mock_client = mock.AsyncMock()
        mock_stop_callback = mock.Mock()
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123).set_stop_callback(mock_stop_callback)
        interval._client = mock_client
        interval._task = mock_task
        interval._tasks = []

        await interval.stop()

        mock_task.cancel.assert_called_once_with()
        assert interval._task is None
        assert interval._tasks == []
        mock_client.call_with_async_di.assert_awaited_once_with(mock_stop_callback)

    @pytest.mark.asyncio()
    async def test_stop_when_stop_callback_stop_raises(self):
        mock_task = mock.Mock()
        mock_client = mock.AsyncMock()
        mock_client.call_with_async_di.side_effect = TypeError
        mock_stop_callback = mock.Mock()
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123).set_stop_callback(mock_stop_callback)
        interval._client = mock_client
        interval._task = mock_task
        interval._add_task(asyncio.create_task(asyncio.sleep(0.1, result=None)))

        await interval.stop()

        mock_task.cancel.assert_called_once_with()
        assert interval._task is None
        assert interval._tasks == []
        mock_client.call_with_async_di.assert_awaited_once_with(mock_stop_callback)

    @pytest.mark.asyncio()
    async def test_stop_when_stop_callback_raises_ignored(self):
        mock_task = mock.Mock()
        mock_client = mock.AsyncMock()
        mock_client.call_with_async_di.side_effect = RuntimeError
        mock_stop_callback = mock.Mock()
        interval = tanjun.schedules.IntervalSchedule(
            mock.Mock(), 123, ignored_exceptions=(RuntimeError,)
        ).set_stop_callback(mock_stop_callback)
        interval._client = mock_client
        interval._task = mock_task
        interval._tasks = []

        await interval.stop()

        mock_task.cancel.assert_called_once_with()
        assert interval._task is None
        assert interval._tasks == []
        mock_client.call_with_async_di.assert_awaited_once_with(mock_stop_callback)

    @pytest.mark.asyncio()
    async def test_stop_when_not_active(self):
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123)

        with pytest.raises(RuntimeError, match="Schedule is not running"):
            await interval.stop()

    def test_with_start_callback(self):
        set_start_callback = mock.Mock()
        interval: tanjun.schedules.IntervalSchedule[typing.Any] = types.new_class(
            "StubIntervalSchedule",
            (tanjun.schedules.IntervalSchedule[typing.Any],),
            exec_body=lambda ns: ns.update({"set_start_callback": set_start_callback}),
        )(mock.Mock(), 123)
        mock_callback = mock.Mock()

        result = interval.with_start_callback(mock_callback)

        assert result is mock_callback
        set_start_callback.assert_called_once_with(mock_callback)

    def test_with_stop_callback(self):
        set_stop_callback = mock.Mock()
        interval: tanjun.schedules.IntervalSchedule[typing.Any] = types.new_class(
            "StubIntervalSchedule",
            (tanjun.schedules.IntervalSchedule[typing.Any],),
            exec_body=lambda ns: ns.update({"set_stop_callback": set_stop_callback}),
        )(mock.Mock(), 123)
        mock_callback = mock.Mock()

        result = interval.with_stop_callback(mock_callback)

        assert result is mock_callback
        set_stop_callback.assert_called_once_with(mock_callback)

    def test_set_ignored_exceptions(self):
        mock_exception: typing.Any = mock.Mock()
        mock_other_exception: typing.Any = mock.Mock()
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123)

        interval.set_ignored_exceptions(mock_exception, mock_other_exception)

        assert interval._ignored_exceptions == (mock_exception, mock_other_exception)

    def test_set_fatal_exceptions(self):
        mock_exception: typing.Any = mock.Mock()
        mock_other_exception: typing.Any = mock.Mock()
        interval = tanjun.schedules.IntervalSchedule(mock.Mock(), 123)

        interval.set_fatal_exceptions(mock_exception, mock_other_exception)

        assert interval._fatal_exceptions == (mock_exception, mock_other_exception)


class TestTimeSchedule:
    def test_callback_property(self):
        mock_callback = mock.AsyncMock()
        interval = tanjun.schedules.TimeSchedule(mock_callback)

        assert interval.callback is mock_callback

    def test_is_alive_property(self):
        interval = tanjun.schedules.TimeSchedule(mock.AsyncMock())

        assert interval.is_alive is False

    @pytest.mark.asyncio()
    async def test_is_alive_property_when_is_alive(self):
        mock_callback = mock.AsyncMock()
        client = alluka.Client()
        interval = tanjun.schedules.TimeSchedule(mock_callback)

        interval.start(client)

        assert interval.is_alive is True

        interval.force_stop()

    @pytest.mark.parametrize(
        ("kwargs", "expected_message"),
        [
            pytest.param(
                {"months": 0},
                r"months value must be \(inclusively\) between 1 and 12, not 0",
                id="Single month too small",
            ),
            pytest.param(
                {"months": 13},
                r"months value must be \(inclusively\) between 1 and 12, not 13",
                id="Single month too large",
            ),
            pytest.param(
                {"months": [-1, 0, 4, 5]},
                r"months must be \(inclusively\) between 1 and 12, not -1 and 5",
                id="Multiple months too small",
            ),
            pytest.param(
                {"months": [4, 5, 7, 14]},
                r"months must be \(inclusively\) between 1 and 12, not 4 and 14",
                id="Multiple months too large",
            ),
            pytest.param(
                {"months": range(0, 14)},
                r"months must be \(inclusively\) between 1 and 12, not 0 and 13",
                id="Months range out of range",
            ),
            pytest.param(
                {"days": 0}, r"days value must be \(inclusively\) between 1 and 31, not 0", id="Single day too small"
            ),
            pytest.param(
                {"days": 32}, r"days value must be \(inclusively\) between 1 and 31, not 32", id="Single day too large"
            ),
            pytest.param(
                {"days": [-1, 0, 4, 5]},
                r"days must be \(inclusively\) between 1 and 31, not -1 and 5",
                id="Multiple days too small",
            ),
            pytest.param(
                {"days": [4, 5, 7, 32]},
                r"days must be \(inclusively\) between 1 and 31, not 4 and 32",
                id="Multiple days too large",
            ),
            pytest.param(
                {"days": range(0, 34)},
                r"days must be \(inclusively\) between 1 and 31, not 0 and 33",
                id="days range out of range",
            ),
            pytest.param(
                {"days": 0, "weekly": True},
                r"days value must be \(inclusively\) between 1 and 8, not 0",
                id="Single day too small and weekly",
            ),
            pytest.param(
                {"days": 9, "weekly": True},
                r"days value must be \(inclusively\) between 1 and 8, not 9",
                id="Single day too large and weekly",
            ),
            pytest.param(
                {"days": [-1, 0, 4, 5], "weekly": True},
                r"days must be \(inclusively\) between 1 and 8, not -1 and 5",
                id="Multiple days too small and weekly",
            ),
            pytest.param(
                {"days": [4, 5, 7, 9], "weekly": True},
                r"days must be \(inclusively\) between 1 and 8, not 4 and 9",
                id="Multiple days too large and weekly",
            ),
            pytest.param(
                {"days": range(1, 10), "weekly": True},
                r"days must be \(inclusively\) between 1 and 8, not 1 and 9",
                id="days range out of range and weekly",
            ),
            pytest.param(
                {"hours": -1},
                r"hours value must be \(inclusively\) between 0 and 23, not -1",
                id="Single hour too small",
            ),
            pytest.param(
                {"hours": 24},
                r"hours value must be \(inclusively\) between 0 and 23, not 24",
                id="Single hour too large",
            ),
            pytest.param(
                {"hours": [-1, 0, 4, 5]},
                r"hours must be \(inclusively\) between 0 and 23, not -1 and 5",
                id="Multiple hours too small",
            ),
            pytest.param(
                {"hours": [4, 5, 7, 25]},
                r"hours must be \(inclusively\) between 0 and 23, not 4 and 25",
                id="Multiple hours too large",
            ),
            pytest.param(
                {"hours": range(0, 25)},
                r"hours must be \(inclusively\) between 0 and 23, not 0 and 24",
                id="Hours range out of range",
            ),
            pytest.param(
                {"minutes": -1},
                r"minutes value must be \(inclusively\) between 0 and 59, not -1",
                id="Single minute too small",
            ),
            pytest.param(
                {"minutes": 60},
                r"minutes value must be \(inclusively\) between 0 and 59, not 60",
                id="Single minute too large",
            ),
            pytest.param(
                {"minutes": [-1, 0, 4, 5]},
                r"minutes must be \(inclusively\) between 0 and 59, not -1 and 5",
                id="Multiple minutes too small",
            ),
            pytest.param(
                {"minutes": [4, 5, 7, 60]},
                r"minutes must be \(inclusively\) between 0 and 59, not 4 and 60",
                id="Multiple minutes too large",
            ),
            pytest.param(
                {"minutes": range(0, 61)},
                r"minutes must be \(inclusively\) between 0 and 59, not 0 and 60",
                id="Minutes range out of range",
            ),
            #
            pytest.param(
                {"seconds": -1},
                r"seconds value must be \(inclusively\) between 0 and 59, not -1",
                id="Single second too small",
            ),
            pytest.param(
                {"seconds": 60},
                r"seconds value must be \(inclusively\) between 0 and 59, not 60",
                id="Single second too large",
            ),
            pytest.param(
                {"seconds": [-1, 0, 4, 5]},
                r"seconds must be \(inclusively\) between 0 and 59, not -1 and 5",
                id="Multiple seconds too small",
            ),
            pytest.param(
                {"seconds": [4, 5, 7, 60]},
                r"seconds must be \(inclusively\) between 0 and 59, not 4 and 60",
                id="Multiple seconds too large",
            ),
            pytest.param(
                {"seconds": range(0, 61)},
                r"seconds must be \(inclusively\) between 0 and 59, not 0 and 60",
                id="Seconds range out of range",
            ),
        ],
    )
    def test_init_with_out_of_range_value(self, kwargs: dict[str, typing.Any], expected_message: str):
        with pytest.raises(ValueError, match=expected_message):
            tanjun.schedules.TimeSchedule(mock.Mock(), **kwargs)

    @pytest.mark.parametrize(
        ("kwargs", "expected_message"),
        [
            pytest.param({"months": 2.4}, "months value must be an integer, not a float", id="Single month"),
            pytest.param({"months": [4, 6, 7, 3.4, 6]}, "Cannot pass floats for months", id="Multiple months"),
            pytest.param({"days": 5.34}, "days value must be an integer, not a float", id="Single day"),
            pytest.param({"days": [3, 5, 2, 4, 5.34]}, "Cannot pass floats for days", id="Multiple days"),
            pytest.param({"hours": 4.3}, "hours value must be an integer, not a float", id="Single hour"),
            pytest.param({"hours": [3, 5, 2, 4.5, 6]}, "Cannot pass floats for hours", id="Multiple hours"),
            pytest.param({"minutes": 3.5}, "minutes value must be an integer, not a float", id="Single minute"),
            pytest.param({"minutes": [4, 5, 5.6, 6, 7]}, "Cannot pass floats for minutes", id="Multiple minutes"),
            pytest.param({"seconds": 3.5}, "seconds value must be an integer, not a float", id="Single second"),
            pytest.param({"seconds": [3, 4, 5, 6.54, 3]}, "Cannot pass floats for seconds", id="Multiple seconds"),
        ],
    )
    def test_init_when_float_passed(self, kwargs: dict[str, typing.Any], expected_message: str):
        with pytest.raises(TypeError, match=expected_message):
            tanjun.schedules.TimeSchedule(mock.Mock(), **kwargs)

    @pytest.mark.asyncio()
    async def test_call_dunder_method(self):
        mock_callback: typing.Any = mock.AsyncMock()
        interval = tanjun.schedules.TimeSchedule(mock_callback)

        result = await interval(123, "32", a=432, b=123)

        assert result is None
        mock_callback.assert_awaited_once_with(123, "32", a=432, b=123)

    def test_copy(self):
        mock_callback: typing.Any = mock.AsyncMock()
        interval = tanjun.schedules.TimeSchedule(mock_callback)
        interval._tasks.append(mock.Mock())

        result = interval.copy()

        assert result is not interval
        assert result.callback is mock_callback
        assert result._tasks == []
        assert result._tasks is not interval._tasks
        assert result._config == interval._config
        assert result._config is not interval._config

    def test_copy_when_schedule_is_active(self):
        interval = tanjun.schedules.TimeSchedule(mock.Mock())
        interval._task = mock.Mock()

        with pytest.raises(RuntimeError, match="Cannot copy an active schedule"):
            interval.copy()

    def test_load_into_component(self):
        mock_component = mock.Mock(tanjun.Component)
        interval = tanjun.schedules.TimeSchedule(mock.AsyncMock())

        interval.load_into_component(mock_component)

        mock_component.add_schedule.assert_called_once_with(interval)

    def test_load_into_component_when_not_loader(self):
        mock_component = mock.Mock(object)
        interval = tanjun.schedules.TimeSchedule(mock.AsyncMock())

        interval.load_into_component(mock_component)

    def test_start(self):
        class StubSchedule(tanjun.schedules.TimeSchedule[typing.Any]):
            ...

        mock_client = mock.Mock()
        interval = StubSchedule(mock.AsyncMock())
        interval._loop = mock.Mock()

        with mock.patch.object(asyncio, "get_running_loop") as get_running_loop:
            interval.start(mock_client)

        assert interval._task is get_running_loop.return_value.create_task.return_value
        assert interval.is_alive is True
        get_running_loop.assert_called_once_with()
        get_running_loop.return_value.create_task.assert_called_once_with(interval._loop.return_value)
        interval._loop.assert_called_once_with(mock_client)

    def test_start_when_passed_event_loop(self):
        class StubSchedule(tanjun.schedules.TimeSchedule[typing.Any]):
            ...

        mock_client = mock.Mock()
        mock_loop = mock.Mock()
        interval = StubSchedule(mock.AsyncMock())
        interval._loop = mock.Mock()

        interval.start(mock_client, loop=mock_loop)

        assert interval._task is mock_loop.create_task.return_value
        assert interval.is_alive is True
        mock_loop.create_task.assert_called_once_with(interval._loop.return_value)
        interval._loop.assert_called_once_with(mock_client)

    def test_start_when_passed_event_loop_isnt_active(self):
        interval = tanjun.schedules.TimeSchedule(mock.AsyncMock())
        mock_loop = mock.Mock()
        mock_loop.is_running.return_value = False

        with pytest.raises(RuntimeError, match="Event loop is not running"):
            interval.start(mock.Mock(), loop=mock_loop)

    @pytest.mark.asyncio()
    async def test_start_when_already_running(self):
        interval = tanjun.schedules.TimeSchedule(mock.AsyncMock())
        interval.start(mock.Mock())
        try:
            with pytest.raises(RuntimeError, match="Schedule is already running"):
                interval.start(mock.Mock())

        finally:
            interval.force_stop()

    @pytest.mark.asyncio()
    async def test_force_stop(self):
        mock_loop_task = mock.Mock()
        mock_task_1 = asyncio.create_task(asyncio.sleep(60, result=None))
        mock_task_2 = asyncio.create_task(asyncio.sleep(60, result=None))
        mock_task_3 = asyncio.create_task(asyncio.sleep(60, result=None))
        interval = tanjun.schedules.TimeSchedule(mock.Mock())
        interval._task = mock_loop_task
        interval._add_task(mock_task_1)
        interval._add_task(mock_task_2)
        interval._add_task(mock_task_3)

        interval.force_stop()
        # This is done to allow any finished tasks to be removed.
        await asyncio.sleep(0)

        assert interval.is_alive is False
        assert interval._task is None
        assert mock_task_1.cancelled() is True
        assert mock_task_2.cancelled() is True
        assert mock_task_3.cancelled() is True

    def test_force_stop_when_no_tasks(self):
        mock_loop_task = mock.Mock()
        interval = tanjun.schedules.TimeSchedule(mock.Mock())
        interval._task = mock_loop_task

        interval.force_stop()

        assert interval.is_alive is False
        assert interval._task is None

    def test_force_stop_when_not_active(self):
        interval = tanjun.schedules.TimeSchedule(mock.Mock())

        with pytest.raises(RuntimeError, match="Schedule is not running"):
            interval.force_stop()

    @pytest.mark.asyncio()
    async def test_stop(self):
        mock_task = mock.Mock()
        mock_task_1 = asyncio.create_task(asyncio.sleep(0.2, result=None))
        mock_task_2 = asyncio.create_task(asyncio.sleep(0.2, result=None))
        mock_task_3 = asyncio.create_task(asyncio.sleep(0.2, result=None))
        interval = tanjun.schedules.TimeSchedule(mock.AsyncMock())
        interval._task = mock_task
        interval._add_task(mock_task_1)
        interval._add_task(mock_task_2)
        interval._add_task(mock_task_3)

        await interval.stop()

        mock_task.cancel.assert_called_once_with()
        assert interval.is_alive is False
        assert interval._task is None
        assert interval._tasks == []
        assert mock_task_1.result() is None
        assert mock_task_2.result() is None
        assert mock_task_3.result() is None

    @pytest.mark.asyncio()
    async def test_stop_when_some_tasks_time_out(self):
        mock_task = mock.Mock()
        mock_task_1 = asyncio.create_task(asyncio.sleep(0.6, result=None))
        mock_task_2 = asyncio.create_task(asyncio.sleep(0.2, result=None))
        mock_task_3 = asyncio.create_task(asyncio.sleep(0.5, result=None))
        mock_task_4 = asyncio.create_task(asyncio.sleep(0.1, result=None))
        interval = tanjun.schedules.TimeSchedule(mock.AsyncMock())
        interval._task = mock_task
        interval._add_task(mock_task_1)
        interval._add_task(mock_task_2)
        interval._add_task(mock_task_3)
        interval._add_task(mock_task_4)

        with pytest.raises(asyncio.TimeoutError):
            await asyncio.wait_for(interval.stop(), 0.25)

        mock_task.cancel.assert_called_once_with()
        assert interval.is_alive is False
        assert interval._task is None
        assert interval._tasks == []
        assert mock_task_1.cancelled() is True
        assert mock_task_2.result() is None
        assert mock_task_3.cancelled() is True
        assert mock_task_4.result() is None

    @pytest.mark.asyncio()
    async def test_stop_when_not_running(self):
        interval = tanjun.schedules.TimeSchedule(mock.AsyncMock())

        with pytest.raises(RuntimeError, match="Schedule is not running"):
            await interval.stop()

    # Note: these have to be at least a microsecond after the target time as
    # the unix event loop won't return the sleep until the target time has passed,
    # not just been reached.
    @pytest.mark.parametrize(
        ("kwargs", "start", "tick_fors", "sleep_for", "expected_dates"),
        [
            pytest.param(
                {"months": 1, "days": 1, "hours": 0, "minutes": 0, "seconds": 0},
                datetime.datetime(2020, 12, 31, 23, 59, 59),
                [
                    datetime.timedelta(seconds=1, microseconds=500001),
                    datetime.timedelta(days=365),
                    datetime.timedelta(days=365),
                ],
                datetime.timedelta(days=730, seconds=2, microseconds=500001),
                [
                    datetime.datetime(2021, 1, 1, 0, 0, 0, 500001),
                    datetime.datetime(2022, 1, 1, 0, 0, 0, 500001),
                    datetime.datetime(2023, 1, 1, 0, 0, 0, 500001),
                ],
                id="Start of each section",
            ),
            pytest.param(
                {"months": 1, "weekly": True, "days": 1, "hours": 0, "minutes": 0, "seconds": 0},
                datetime.datetime(2066, 12, 31, 23, 59, 59),
                [
                    datetime.timedelta(days=2, seconds=1, microseconds=500001),
                    *(datetime.timedelta(days=7),) * 4,
                    datetime.timedelta(days=336),
                    *(datetime.timedelta(days=7),) * 4,
                ],
                datetime.timedelta(days=394, seconds=2, microseconds=500001),
                [
                    datetime.datetime(2067, 1, 3, 0, 0, 0, 500001),
                    datetime.datetime(2067, 1, 10, 0, 0, 0, 500001),
                    datetime.datetime(2067, 1, 17, 0, 0, 0, 500001),
                    datetime.datetime(2067, 1, 24, 0, 0, 0, 500001),
                    datetime.datetime(2067, 1, 31, 0, 0, 0, 500001),
                    datetime.datetime(2068, 1, 2, 0, 0, 0, 500001),
                    datetime.datetime(2068, 1, 9, 0, 0, 0, 500001),
                    datetime.datetime(2068, 1, 16, 0, 0, 0, 500001),
                    datetime.datetime(2068, 1, 23, 0, 0, 0, 500001),
                    datetime.datetime(2068, 1, 30, 0, 0, 0, 500001),
                ],
                id="Start of each section weekly",
            ),
            # TODO: End of each section  see histpry for test data
            # TODO: End of each section weekly  see history for test data
            pytest.param(
                {},
                datetime.datetime(2011, 1, 4, 11, 57, 10),
                [datetime.timedelta(seconds=50, milliseconds=500, microseconds=1), datetime.timedelta(minutes=1)],
                datetime.timedelta(minutes=6, seconds=30),
                [
                    datetime.datetime(2011, 1, 4, 11, 58, 0, 500001),
                    datetime.datetime(2011, 1, 4, 11, 59, 0, 500001),
                    datetime.datetime(2011, 1, 4, 12, 0, 0, 500001),
                    datetime.datetime(2011, 1, 4, 12, 1, 0, 500001),
                    datetime.datetime(2011, 1, 4, 12, 2, 0, 500001),
                    datetime.datetime(2011, 1, 4, 12, 3, 0, 500001),
                ],
                id="default timing",
            ),
            pytest.param(
                {},
                datetime.datetime(2035, 11, 3, 23, 56, 5),
                [datetime.timedelta(seconds=55, milliseconds=500, microseconds=1), datetime.timedelta(minutes=1)],
                datetime.timedelta(minutes=6, seconds=30),
                [
                    datetime.datetime(2035, 11, 3, 23, 57, 0, 500001),
                    datetime.datetime(2035, 11, 3, 23, 58, 0, 500001),
                    datetime.datetime(2035, 11, 3, 23, 59, 0, 500001),
                    datetime.datetime(2035, 11, 4, 0, 0, 0, 500001),
                    datetime.datetime(2035, 11, 4, 0, 1, 0, 500001),
                    datetime.datetime(2035, 11, 4, 0, 2, 0, 500001),
                ],
                id="default timing bumps year",
            ),
            pytest.param(
                {"months": [7, 4], "days": [14, 7], "hours": [17, 12], "minutes": [55, 22], "seconds": [30, 10]},
                datetime.datetime(2016, 3, 4, 10, 40, 30),
                _chain(
                    (
                        d,
                        datetime.timedelta(seconds=20),
                        datetime.timedelta(minutes=32, seconds=40),
                        datetime.timedelta(seconds=20),
                    )
                    for d in (
                        datetime.timedelta(days=34, seconds=6100, microseconds=500001),
                        datetime.timedelta(hours=4, minutes=26, seconds=40),
                        datetime.timedelta(days=6, hours=18, minutes=26, seconds=40),
                        datetime.timedelta(hours=4, minutes=26, seconds=40),
                        datetime.timedelta(days=83, hours=18, minutes=26, seconds=40),
                        datetime.timedelta(hours=4, minutes=26, seconds=40),
                        datetime.timedelta(days=6, hours=18, minutes=26, seconds=40),
                        datetime.timedelta(hours=4, minutes=26, seconds=40),
                    )
                )
                + [
                    datetime.timedelta(days=266, hours=18, minutes=26, seconds=40),
                    datetime.timedelta(seconds=20),
                    datetime.timedelta(seconds=1),
                ],
                datetime.timedelta(days=399, hours=1, minutes=42, seconds=1),
                [
                    *(
                        datetime.datetime(2016, *args, microsecond=500001)
                        for args in itertools.product([4, 7], [7, 14], [12, 17], [22, 55], [10, 30])
                    ),
                    datetime.datetime(2017, 4, 7, 12, 22, 10, 500001),
                    datetime.datetime(2017, 4, 7, 12, 22, 30, 500001),
                ],
                id="all time fields specified",
            ),
            pytest.param(
                {"months": [1, 5, 9], "weekly": True, "days": [3, 5], "hours": [5], "minutes": [45], "seconds": [10]},
                datetime.datetime(2069, 3, 5, 5, 45, 10),
                [
                    datetime.timedelta(days=57, microseconds=500001),
                    *(datetime.timedelta(days=2), datetime.timedelta(days=5)) * 4,
                    datetime.timedelta(days=2),
                    datetime.timedelta(days=96),
                    *(datetime.timedelta(days=2), datetime.timedelta(days=5)) * 3,
                    datetime.timedelta(days=2),
                    datetime.timedelta(days=96),
                    *(datetime.timedelta(days=2), datetime.timedelta(days=5)) * 4,
                    datetime.timedelta(days=2),
                ],
                datetime.timedelta(days=333),
                [
                    datetime.datetime(2069, 5, 1, 5, 45, 10, 500001),  # 3rd day
                    datetime.datetime(2069, 5, 3, 5, 45, 10, 500001),
                    datetime.datetime(2069, 5, 8, 5, 45, 10, 500001),
                    datetime.datetime(2069, 5, 10, 5, 45, 10, 500001),
                    datetime.datetime(2069, 5, 15, 5, 45, 10, 500001),
                    datetime.datetime(2069, 5, 17, 5, 45, 10, 500001),
                    datetime.datetime(2069, 5, 22, 5, 45, 10, 500001),
                    datetime.datetime(2069, 5, 24, 5, 45, 10, 500001),
                    datetime.datetime(2069, 5, 29, 5, 45, 10, 500001),
                    datetime.datetime(2069, 5, 31, 5, 45, 10, 500001),
                    datetime.datetime(2069, 9, 4, 5, 45, 10, 500001),  # 3rd day
                    datetime.datetime(2069, 9, 6, 5, 45, 10, 500001),
                    datetime.datetime(2069, 9, 11, 5, 45, 10, 500001),
                    datetime.datetime(2069, 9, 13, 5, 45, 10, 500001),
                    datetime.datetime(2069, 9, 18, 5, 45, 10, 500001),
                    datetime.datetime(2069, 9, 20, 5, 45, 10, 500001),
                    datetime.datetime(2069, 9, 25, 5, 45, 10, 500001),
                    datetime.datetime(2069, 9, 27, 5, 45, 10, 500001),
                    datetime.datetime(2070, 1, 1, 5, 45, 10, 500001),  # 3rd day
                    datetime.datetime(2070, 1, 3, 5, 45, 10, 500001),
                    datetime.datetime(2070, 1, 8, 5, 45, 10, 500001),
                    datetime.datetime(2070, 1, 10, 5, 45, 10, 500001),
                    datetime.datetime(2070, 1, 15, 5, 45, 10, 500001),
                    datetime.datetime(2070, 1, 17, 5, 45, 10, 500001),
                    datetime.datetime(2070, 1, 22, 5, 45, 10, 500001),
                    datetime.datetime(2070, 1, 24, 5, 45, 10, 500001),
                    datetime.datetime(2070, 1, 29, 5, 45, 10, 500001),
                    datetime.datetime(2070, 1, 31, 5, 45, 10, 500001),
                ],
                id="all time fields specified weekly",
            ),
            pytest.param(
                {"months": range(11, 13), "days": [1, 15], "hours": range(7, 5, -1), "minutes": range(3, 1, -1)},
                datetime.datetime(2016, 7, 15, 12, 22, 10, 500001),
                _chain(
                    (d, datetime.timedelta(minutes=1), datetime.timedelta(minutes=59), datetime.timedelta(minutes=1))
                    for d in (
                        datetime.timedelta(days=108, hours=17, minutes=39, seconds=50),
                        datetime.timedelta(days=13, hours=22, minutes=59),
                        datetime.timedelta(days=15, hours=22, minutes=59),
                        datetime.timedelta(days=13, hours=22, minutes=59),
                        datetime.timedelta(days=320, hours=22, minutes=59),
                        datetime.timedelta(days=13, hours=22, minutes=59),
                    )
                )
                + [datetime.timedelta(days=3)],
                datetime.timedelta(days=487, hours=18, minutes=41, seconds=50),
                [
                    *(
                        datetime.datetime(2016, *args, microsecond=500001)
                        for args in itertools.product([11, 12], [1, 15], [6, 7], [2, 3])
                    ),
                    *(
                        datetime.datetime(2017, 11, *args, microsecond=500001)
                        for args in itertools.product([1, 15], [6, 7], [2, 3])
                    ),
                ],
                id="specific months, days, hours and minutes",
            ),
            pytest.param(
                {
                    "months": range(12, 10, -1),
                    "weekly": True,
                    "days": range(3, 5),
                    "hours": range(6, 8),
                    "minutes": range(3, 4),
                },
                datetime.datetime(2019, 5, 1),
                [
                    datetime.timedelta(days=189, hours=6, minutes=3, microseconds=500001),
                    datetime.timedelta(hours=1),
                    datetime.timedelta(hours=23),
                    datetime.timedelta(hours=1),
                    *(
                        datetime.timedelta(days=5, hours=23),
                        datetime.timedelta(hours=1),
                        datetime.timedelta(hours=23),
                        datetime.timedelta(hours=1),
                    )
                    * 7,
                    datetime.timedelta(days=313, hours=23),
                    datetime.timedelta(hours=1),
                    datetime.timedelta(hours=23),
                ],
                datetime.timedelta(days=554, hours=6, minutes=4),
                [
                    datetime.datetime(2019, 11, 6, 6, 3, 00, 500001),
                    datetime.datetime(2019, 11, 6, 7, 3, 00, 500001),
                    datetime.datetime(2019, 11, 7, 6, 3, 00, 500001),
                    datetime.datetime(2019, 11, 7, 7, 3, 00, 500001),
                    datetime.datetime(2019, 11, 13, 6, 3, 00, 500001),
                    datetime.datetime(2019, 11, 13, 7, 3, 00, 500001),
                    datetime.datetime(2019, 11, 14, 6, 3, 00, 500001),
                    datetime.datetime(2019, 11, 14, 7, 3, 00, 500001),
                    datetime.datetime(2019, 11, 20, 6, 3, 00, 500001),
                    datetime.datetime(2019, 11, 20, 7, 3, 00, 500001),
                    datetime.datetime(2019, 11, 21, 6, 3, 00, 500001),
                    datetime.datetime(2019, 11, 21, 7, 3, 00, 500001),
                    datetime.datetime(2019, 11, 27, 6, 3, 00, 500001),
                    datetime.datetime(2019, 11, 27, 7, 3, 00, 500001),
                    datetime.datetime(2019, 11, 28, 6, 3, 00, 500001),
                    datetime.datetime(2019, 11, 28, 7, 3, 00, 500001),
                    datetime.datetime(2019, 12, 4, 6, 3, 00, 500001),
                    datetime.datetime(2019, 12, 4, 7, 3, 00, 500001),
                    datetime.datetime(2019, 12, 5, 6, 3, 00, 500001),
                    datetime.datetime(2019, 12, 5, 7, 3, 00, 500001),
                    datetime.datetime(2019, 12, 11, 6, 3, 00, 500001),
                    datetime.datetime(2019, 12, 11, 7, 3, 00, 500001),
                    datetime.datetime(2019, 12, 12, 6, 3, 00, 500001),
                    datetime.datetime(2019, 12, 12, 7, 3, 00, 500001),
                    datetime.datetime(2019, 12, 18, 6, 3, 00, 500001),
                    datetime.datetime(2019, 12, 18, 7, 3, 00, 500001),
                    datetime.datetime(2019, 12, 19, 6, 3, 00, 500001),
                    datetime.datetime(2019, 12, 19, 7, 3, 00, 500001),
                    datetime.datetime(2019, 12, 25, 6, 3, 00, 500001),
                    datetime.datetime(2019, 12, 25, 7, 3, 00, 500001),
                    datetime.datetime(2019, 12, 26, 6, 3, 00, 500001),
                    datetime.datetime(2019, 12, 26, 7, 3, 00, 500001),
                    datetime.datetime(2020, 11, 4, 6, 3, 00, 500001),
                    datetime.datetime(2020, 11, 4, 7, 3, 00, 500001),
                    datetime.datetime(2020, 11, 5, 6, 3, 00, 500001),
                ],
                id="specific months, days, hours and minutes weekly",
            ),
            # TODO: specific months, days, hours and seconds  See history for test data
            # kwargs, start, tick_fors, sleep_for, expected_dates  TODO: test timezone behaviour
            # needs to be singled: days, days weekly, minutes, seconds
            # TODO: specific months, days, hours and seconds weekly  #  backwards range days and seconds
            # TODO: specific months, days, minutes and seconds   # backwards range days
            # TODO: specific months, days, minutes and seconds weekly
            # TODO: specific months, hours, minutes and seconds
            # TODO: specific months, days and hours
            # TODO: specific months, days and hours weekly
            # TODO: specific months, days and minutes
            # TODO: specific months, days and minutes weekly
            # TODO: specific months, days and seconds
            # TODO: specific months, days and seconds weekly
            # TODO: specific months, hours and minutes
            # TODO: specific months, hours and seconds
            # TODO: specific months, minutes and seconds
            # TODO: specific months and days
            # TODO: specific months and days weekly
            # TODO: specific months and hours
            # TODO: specific months and minutes
            # TODO: specific months and seconds
            # TODO: specific months
            # TODO: specific days, hours minutes and seconds
            # TODO: specific days, hours minutes and seconds weekly
            # TODO: specific days, hours and minutes
            # TODO: specific days, hours and minutes weekly
            # TODO: specific days, hours and seconds
            # TODO: specific days, hours and seconds weekly
            # TODO: specific days, minutes and seconds
            # TODO: specific days, minutes and seconds weekly
            # TODO: specific days and hours
            # TODO: specific days and hours weekly
            # TODO: specific days and minutes
            # TODO: specific days and minutes weekly
            # TODO: specific days and seconds
            # TODO: specific days and seconds weekly
            # TODO: specific days
            # TODO: specific days weekly
            # TODO: specific hours, minutes and seconds
            # TODO: specific hours and minutes
            # TODO: specific hours and seconds
            # TODO: specific hours
            # TODO: specific minutes and seconds
            # TODO: specific minutes
            # TODO: specific seconds
        ],
    )
    @pytest.mark.timeout(_TIMEOUT)
    @pytest.mark.asyncio()
    async def test_run(
        self,
        kwargs: dict[str, typing.Any],
        start: datetime.datetime,
        tick_fors: list[datetime.timedelta],
        sleep_for: datetime.timedelta,
        expected_dates: list[datetime.datetime],
    ):
        # Ensure test-data integrity
        assert start < expected_dates[0], "Start must be before expected dates"
        assert sorted(expected_dates) == expected_dates, "Expected dates must be sorted"

        called_at: list[datetime.datetime] = []

        @_print_tb
        async def callback():
            called_at.append(datetime.datetime.now())
            clock.spawn_ticker()

        schedule = tanjun.schedules.as_time_schedule(**kwargs)(callback)

        with freezegun.freeze_time(start, tick=False) as frozen_time:
            assert isinstance(frozen_time, freezegun.api.FrozenDateTimeFactory)
            clock = _ManualClock(frozen_time, tick_fors)

            schedule.start(alluka.Client())
            clock.spawn_ticker()
            await asyncio.sleep(sleep_for.total_seconds())
            await schedule.stop()
            clock.stop_ticker()

        assert called_at == expected_dates

    @pytest.mark.timeout(_TIMEOUT)
    @pytest.mark.asyncio()
    async def test_error_handling(self):
        called_at: list[datetime.datetime] = []

        @_print_tb
        async def callback() -> None:
            called_at.append(datetime.datetime.now())
            clock.spawn_ticker()
            length = len(called_at)
            if length == 1:
                raise RuntimeError("Not caught")

            if length == 2:
                raise ValueError("Ignored")

            if length == 3:
                raise TypeError("Fatal")

        schedule = (
            tanjun.schedules.as_time_schedule(hours=[4, 6], minutes=30)(callback)
            .set_fatal_exceptions(TypeError)
            .set_ignored_exceptions(ValueError)
        )

        with freezegun.freeze_time(datetime.datetime(2044, 4, 4), tick=False) as frozen_time:
            assert isinstance(frozen_time, freezegun.api.FrozenDateTimeFactory)
            clock = _ManualClock(
                frozen_time,
                [
                    datetime.timedelta(hours=4, minutes=30, microseconds=500001),
                    datetime.timedelta(hours=2),
                    datetime.timedelta(hours=22),
                ],
            ).spawn_ticker()
            schedule.start(alluka.Client())
            await asyncio.sleep(datetime.timedelta(days=1, seconds=16201).total_seconds())

            clock.stop_ticker()
            assert schedule.is_alive is False

        assert called_at == [
            datetime.datetime(2044, 4, 4, 4, 30, 0, 500001),
            datetime.datetime(2044, 4, 4, 6, 30, 0, 500001),
            datetime.datetime(2044, 4, 5, 4, 30, 0, 500001),
        ]



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pyright: reportIncompatibleMethodOverride=none
# pyright: reportUnknownMemberType=none
# This leads to too many false-positives around mocks.

from unittest import mock

import hikari
import pytest

import tanjun


class TestCommandError:
    def test_init_dunder_method(self):
        mock_attachment = mock.Mock()
        mock_component = mock.Mock()
        mock_embed = mock.Mock()

        error = tanjun.CommandError(
            "moon",
            delete_after=555,
            attachments=[mock_attachment],
            components=[mock_component],
            embeds=[mock_embed],
            mentions_everyone=False,
            user_mentions=[2332],
            role_mentions=[6534],
        )

        assert error.content == "moon"
        assert error.attachments == [mock_attachment]
        assert error.components == [mock_component]
        assert error.delete_after == 555
        assert error.embeds == [mock_embed]
        assert error.mentions_everyone is False
        assert error.user_mentions == [2332]
        assert error.role_mentions == [6534]

    def test_init_dunder_method_for_singular_fields(self):
        mock_attachment = mock.Mock()
        mock_component = mock.Mock()
        mock_embed = mock.Mock()

        error = tanjun.CommandError(attachment=mock_attachment, component=mock_component, embed=mock_embed)

        assert error.attachments == [mock_attachment]
        assert error.components == [mock_component]
        assert error.embeds == [mock_embed]

    def test_init_dunder_method_for_partial(self):
        error = tanjun.CommandError()

        assert error.content is hikari.UNDEFINED
        assert error.attachments is hikari.UNDEFINED
        assert error.components is hikari.UNDEFINED
        assert error.delete_after is None
        assert error.embeds is hikari.UNDEFINED
        assert error.mentions_everyone is hikari.UNDEFINED
        assert error.role_mentions is hikari.UNDEFINED
        assert error.user_mentions is hikari.UNDEFINED

    def test_init_dunder_method_when_both_attachment_and_attachments_passed(self):
        with pytest.raises(ValueError, match="Cannot specify both attachment and attachments"):
            tanjun.CommandError(attachment=mock.Mock(), attachments=[mock.Mock()])

    def test_init_dunder_method_when_both_component_and_components_passed(self):
        with pytest.raises(ValueError, match="Cannot specify both component and components"):
            tanjun.CommandError(component=mock.Mock(), components=[mock.Mock()])

    def test_init_dunder_method_when_both_embed_and_embeds_passed(self):
        with pytest.raises(ValueError, match="Cannot specify both embed and embeds"):
            tanjun.CommandError(embed=mock.Mock(), embeds=[mock.Mock()])

    def test_str_dunder_method(self):
        assert str(tanjun.CommandError("bar")) == "bar"

    @pytest.mark.asyncio()
    async def test_send(self):
        error = tanjun.CommandError()
        mock_context = mock.AsyncMock()

        result = await error.send(mock_context)

        assert result is mock_context.respond.return_value
        mock_context.respond.assert_awaited_once_with(
            content=hikari.UNDEFINED,
            attachments=hikari.UNDEFINED,
            components=hikari.UNDEFINED,
            delete_after=None,
            embeds=hikari.UNDEFINED,
            ensure_result=False,
            mentions_everyone=hikari.UNDEFINED,
            role_mentions=hikari.UNDEFINED,
            user_mentions=hikari.UNDEFINED,
        )

    @pytest.mark.asyncio()
    async def test_send_when_all_fields(self):
        mock_attachment = mock.Mock()
        mock_component = mock.Mock()
        mock_embed = mock.Mock()
        error = tanjun.CommandError(
            "hello",
            attachments=[mock_attachment],
            components=[mock_component],
            delete_after=53,
            embeds=[mock_embed],
            mentions_everyone=True,
            role_mentions=[123, 431],
            user_mentions=[666, 555],
        )
        mock_context = mock.AsyncMock()

        result = await error.send(mock_context, ensure_result=True)

        assert result is mock_context.respond.return_value
        mock_context.respond.assert_awaited_once_with(
            content="hello",
            attachments=[mock_attachment],
            components=[mock_component],
            delete_after=53,
            embeds=[mock_embed],
            ensure_result=True,
            mentions_everyone=True,
            role_mentions=[123, 431],
            user_mentions=[666, 555],
        )


class TestParserError:
    def test__init__(self):
        error = tanjun.ParserError("bank", "no u")

        assert error.message == "bank"
        assert error.parameter == "no u"

    def test__str__(self):
        assert str(tanjun.ParserError("bankette", "now2")) == "bankette"


class TestConversionError:
    def test__init__(self):
        mock_error = mock.Mock()

        error = tanjun.ConversionError("bankettete", "aye", errors=[mock_error])

        assert error.message == "bankettete"
        assert error.parameter == "aye"
        assert error.errors == (mock_error,)


class TestNotEnoughArgumentsError:
    def test__init__(self):
        error = tanjun.NotEnoughArgumentsError("aye", "naye")

        assert error.message == "aye"
        assert error.parameter == "naye"


class TestTooManyArgumentsError:
    def test__init__(self):
        error = tanjun.TooManyArgumentsError("blank", "fama")

        assert error.message == "blank"
        assert error.parameter == "fama"


class TestModuleMissingLoaders:
    def test___init__(self):
        error = tanjun.ModuleMissingLoaders("foo", "bar")

        assert error.message == "foo"
        assert error.path == "bar"


class TestModuleMissingUnloaders:
    def test___init__(self):
        error = tanjun.ModuleMissingUnloaders("beep", "boop")

        assert error.message == "beep"
        assert error.path == "boop"


class TestModuleStateConflict:
    def test___init__(self):
        error = tanjun.ModuleStateConflict("esxd", "dsaasd")

        assert error.message == "esxd"
        assert error.path == "dsaasd"


class TestFailedModuleLoad:
    def test___init__(self):
        error = tanjun.FailedModuleLoad("beat/my/boobs")

        assert error.path == "beat/my/boobs"


class TestFailedModuleUnload:
    def test___init__(self):
        error = tanjun.FailedModuleUnload("yeet/ok")

        assert error.path == "yeet/ok"



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from unittest import mock

import hikari

import tanjun


class TestBasicLocaliser:
    def test_add_to_client(self):
        client = tanjun.Client(mock.AsyncMock())
        localiser = tanjun.dependencies.BasicLocaliser()

        assert localiser.add_to_client(client) is None

        assert client.injector.get_type_dependency(tanjun.dependencies.AbstractLocaliser) is localiser
        assert client.injector.get_type_dependency(tanjun.dependencies.AbstractLocalizer) is localiser

    def test_get_all_variants(self):
        localiser = tanjun.dependencies.BasicLocaliser().set_variants(
            "yeet",
            {hikari.Locale.BG: "BIGGIE CHEESE", hikari.Locale.CS: "COMPUTER SENTIENT", hikari.Locale.EN_GB: "no"},
            DE="yeet",
            boom="NO",
        )

        assert localiser.get_all_variants("yeet") == {
            hikari.Locale.BG: "BIGGIE CHEESE",
            hikari.Locale.CS: "COMPUTER SENTIENT",
            hikari.Locale.EN_GB: "no",
            "DE": "yeet",
            "boom": "NO",
        }

    def test_get_all_variants_ignores_default(self):
        localiser = tanjun.dependencies.BasicLocaliser().set_variants(
            "ayanami",
            {hikari.Locale.ES_ES: "que?", hikari.Locale.DA: "Danken", "default": "yeet", hikari.Locale.FR: "for real"},
        )

        assert localiser.get_all_variants("ayanami") == {
            hikari.Locale.ES_ES: "que?",
            hikari.Locale.DA: "Danken",
            hikari.Locale.FR: "for real",
        }

    def test_get_all_variants_when_format(self):
        localiser = tanjun.dependencies.BasicLocaliser().set_variants(
            "rei",
            {
                hikari.Locale.DA: "ok {no} meow",
                hikari.Locale.CS: "hikari {meow} nom",
                hikari.Locale.EN_GB: "{no} {meow} uwu me pws",
                hikari.Locale.JA: "bye bye bye",
            },
        )

        assert localiser.get_all_variants("rei", no="123321", meow="6969") == {
            hikari.Locale.DA: "ok 123321 meow",
            hikari.Locale.CS: "hikari 6969 nom",
            hikari.Locale.EN_GB: "123321 6969 uwu me pws",
            hikari.Locale.JA: "bye bye bye",
        }

    def test_get_all_variants_when_not_known(self):
        assert tanjun.dependencies.BasicLocaliser().get_all_variants("nope") == {}

    def test_localise(self):
        localiser = tanjun.dependencies.BasicLocaliser().set_variants(
            "the end", {hikari.Locale.FI: "bye"}, vi="echo", fi="meow"
        )

        assert localiser.localise("the end", "vi") == "echo"

    def test_localise_when_format(self):
        localiser = tanjun.dependencies.BasicLocaliser().set_variants(
            "ANDER",
            {hikari.Locale.EN_GB: "nyaa", hikari.Locale.FR: "ashley {now}", hikari.Locale.IT: "op"},
            moomin="okokokok",
        )

        assert localiser.localise("ANDER", hikari.Locale.FR, now="444") == "ashley 444"

    def test_localise_when_tag_unknown(self):
        localiser = tanjun.dependencies.BasicLocaliser().set_variants("NERV", {hikari.Locale.BG: "nom"})

        assert localiser.localise("NERV", hikari.Locale.EN_GB) is None

    def test_localise_when_id_unknown(self):
        assert tanjun.dependencies.BasicLocaliser().localise("foo", hikari.Locale.FR) is None

    def test_localize(self):
        localiser = tanjun.dependencies.BasicLocaliser().set_variants(
            "the end", {hikari.Locale.FI: "bye"}, vi="echo", fi="meow"
        )

        assert localiser.localize("the end", "vi") == "echo"

    def test_localize_when_format(self):
        localiser = tanjun.dependencies.BasicLocaliser().set_variants(
            "ANDER",
            {hikari.Locale.EN_GB: "nyaa", hikari.Locale.FR: "ashley {now}", hikari.Locale.IT: "op"},
            moomin="okokokok",
        )

        assert localiser.localize("ANDER", hikari.Locale.FR, now="444") == "ashley 444"

    def test_localize_when_tag_unknown(self):
        localiser = tanjun.dependencies.BasicLocaliser().set_variants("NERV", {hikari.Locale.BG: "nom"})

        assert localiser.localize("NERV", hikari.Locale.EN_GB) is None

    def test_localize_when_id_unknown(self):
        assert tanjun.dependencies.BasicLocaliser().localize("foo", hikari.Locale.FR) is None



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
r"""Interface for an optional asynchronous gateway cache dependency.

This allows you to share data between instances using something like a redis cache
(for example) and will be used by standard Tanjun components as well as
extensions if implemented.

!!! note
    While there aren't any standard implementations for these interfaces, a
    Redis implementation of this for the types found in Hikari's gateway cache
    can be found in [hikari-sake](https://github.com/FasterSpeeding/Sake)
    \>=v1.0.1a1 (exposed by [sake.redis.ResourceClient.add_to_tanjun][]).

Tanjun will use the following type dependencies for these interfaces if they are
registered with the client:

* `AsyncCache[str, hikari.InviteWithMetadata]`
* `SfCache[hikari.PermissibleGuildChannel]`
* `SfCache[hikari.GuildThreadChannel]`
* `SfCache[hikari.KnownCustomEmoji]`
* `SfCache[hikari.Guild]`
* `SfCache[hikari.Role]`
* `SfCache[hikari.User]`
* `SfGuildBound[hikari.Member]`
* `SfGuildBound[hikari.MemberPresence]`
* `SfGuildBound[hikari.VoiceState]`
* `SfGuildBound[hikari.Role]`
* `SingleStoreCache[hikari.OwnUser]`
* `SingleStoreCache[hikari.Application]`
* `SingleStoreCache[hikari.AuthorizationApplication]`
"""
from __future__ import annotations

__all__: list[str] = [
    "AsyncCache",
    "CacheIterator",
    "CacheMissError",
    "ChannelBoundCache",
    "EntryNotFound",
    "GuildBoundCache",
    "SfCache",
    "SfChannelBound",
    "SfGuildBound",
    "SingleStoreCache",
]

import abc
import typing

import hikari

from .. import errors

if typing.TYPE_CHECKING:
    _DefaultT = typing.TypeVar("_DefaultT")


_KeyT = typing.TypeVar("_KeyT")
_ValueT = typing.TypeVar("_ValueT")


class CacheMissError(errors.TanjunError):
    """Raised when an entry isn't found in the cache.

    !!! note
        [EntryNotFound][tanjun.dependencies.EntryNotFound] inherits from this
        error and will only be raised if the cache knows that the entry
        doesn't exist.
    """


class EntryNotFound(CacheMissError):
    """Raised when an entry does not exist.

    !!! note
        This is a specialisation of [CacheMissError][tanjun.dependencies.CacheMissError]
        which indicates that the cache is sure that the entry doesn't exist.
    """


class CacheIterator(hikari.LazyIterator[_ValueT]):
    """Abstract interface of a cache resource asynchronous iterator.

    For more information on how this is used, see the documentation for
    [hikari.iterators.LazyIterator][].
    """

    __slots__ = ()

    @abc.abstractmethod
    async def len(self) -> int:
        """Get the length of the target resource.

        !!! note
            Unlike [tanjun.dependencies.CacheIterator.count][], this method
            will not deplete the iterator.

        Returns
        -------
        int
            The length of the targeted resource.
        """


class SingleStoreCache(abc.ABC, typing.Generic[_ValueT]):
    """Abstract interface of a cache which stores one resource.

    !!! note
        This is mostly just for the [hikari.users.OwnUser][] cache store.
    """

    __slots__ = ()

    @abc.abstractmethod
    async def get(self, *, default: _DefaultT = ...) -> typing.Union[_ValueT, _DefaultT]:
        """Get the entry.

        Parameters
        ----------
        default
            The default value to return if an entry wasn't found.

            If provided then no errors will be raised when no entry is found.

        Returns
        -------
        _ValueT | _DefaultT
            The found entry or the default if any was provided.

        Raises
        ------
        CacheMissError
            If the entry wasn't found.

            This won't be raised if `default` is passed.
        EntryNotFound
            If the entry wasn't found and the the entry definitely doesn't exist.

            This won't be raised if `default` is passed.

            This is a specialisation of `CacheMissError` and thus may be
            caught as `CacheMissError and otherwise would need to be before
            `CacheMissError` in a try, multiple catch statement.
        """


class AsyncCache(abc.ABC, typing.Generic[_KeyT, _ValueT]):
    """Abstract interface of a cache which stores globally identifiable resources.

    !!! note
        This will never be implemented for resources such as [hikari.guilds.Member][]
        and [hikari.presences.MemberPresence][] which are only unique per-parent resource.
    """

    __slots__ = ()

    @abc.abstractmethod
    async def get(self, key: _KeyT, /, *, default: _DefaultT = ...) -> typing.Union[_ValueT, _DefaultT]:
        """Get an entry from this cache by ID.

        Parameters
        ----------
        key
            Unique key of the entry to get; this will often be a snowflake.
        default
            The default value to return if an entry wasn't found.

            If provided then no errors will be raised when no entry is found.

        Returns
        -------
        _ValueT | _DefaultT
            The found entry or the default if any was provided.

        Raises
        ------
        CacheMissError
            If the entry wasn't found.

            This won't be raised if `default` is passed.
        EntryNotFound
            If the entry wasn't found and the the entry definitely doesn't exist.

            This won't be raised if `default` is passed.

            This is a specialisation of `CacheMissError` and thus may be
            caught as `CacheMissError and otherwise would need to be before
            `CacheMissError` in a try, multiple catch statement.
        """

    @abc.abstractmethod
    def iter_all(self) -> CacheIterator[_ValueT]:
        """Asynchronously iterate over the globally cached entries for this resource.

        !!! note
            For more information on how this is used, see the documentation for
            [hikari.iterators.LazyIterator][].

        Returns
        -------
        CacheIterator[_ValueT]
            An asynchronous iterator of the entries cached globally for this resource.
        """


SfCache = AsyncCache[hikari.Snowflakeish, _ValueT]
"""Alias of [tanjun.dependencies.AsyncCache][] where the key is a snowflake."""


class ChannelBoundCache(abc.ABC, typing.Generic[_KeyT, _ValueT]):
    """Abstract interface of a cache which stores channel-bound resources."""

    __slots__ = ()

    @abc.abstractmethod
    async def get_from_channel(
        self, channel_id: hikari.Snowflakeish, key: _KeyT, /, *, default: _DefaultT = ...
    ) -> typing.Union[_ValueT, _DefaultT]:
        """Get an entry from this cache for a specific channel by ID.

        Parameters
        ----------
        channel_id
            ID of the channel to get an entry for.
        key
            Unique key of the entry to get; this will usually be a snowflake.
        default
            The default value to return if an entry wasn't found.

            If provided then no errors will be raised when no entry is found.

        Returns
        -------
        _ValueT | _DefaultT
            The found entry or the default if any was provided.

        Raises
        ------
        CacheMissError
            If the entry wasn't found.

            This won't be raised if `default` is passed.
        EntryNotFound
            If the entry wasn't found and the the entry definitely doesn't exist.

            This won't be raised if `default` is passed.

            This is a specialisation of `CacheMissError` and thus may be
            caught as `CacheMissError and otherwise would need to be before
            `CacheMissError` in a try, multiple catch statement.
        """

    @abc.abstractmethod
    def iter_for_channel(self, channel_id: hikari.Snowflakeish, /) -> CacheIterator[_ValueT]:
        """Asynchronously iterate over the entries entries cached for a channel.

        Parameters
        ----------
        channel_id
            ID of the channel to iterate over the entries cached for.

        Returns
        -------
        CacheIterator[_ValueT]
            An asynchronous iterator of the entries cached for the specified channel.
        """

    @abc.abstractmethod
    def iter_all(self) -> CacheIterator[_ValueT]:
        """Asynchronously iterate over the globally cached entries for this resource.

        !!! note
            For more information on how this is used, see the documentation for
            [hikari.iterators.LazyIterator][].

        Returns
        -------
        CacheIterator[_ValueT]
            An asynchronous iterator of the entries cached globally for this resource.
        """


SfChannelBound = ChannelBoundCache[hikari.Snowflakeish, _ValueT]
"""Alias of [tanjun.dependencies.ChannelBoundCache][] where the key is a snowflake."""


class GuildBoundCache(abc.ABC, typing.Generic[_KeyT, _ValueT]):
    """Abstract interface of a cache which stores guild-bound resources."""

    __slots__ = ()

    @abc.abstractmethod
    async def get_from_guild(
        self, guild_id: hikari.Snowflakeish, key: _KeyT, /, *, default: _DefaultT = ...
    ) -> typing.Union[_ValueT, _DefaultT]:
        """Get an entry from this cache for a specific guild by ID.

        Parameters
        ----------
        guild_id
            ID of the guild to get an entry for.
        key
            Unique key of the entry to get; this will usually be a snowflake.
        default
            The default value to return if an entry wasn't found.

            If provided then no errors will be raised when no entry is found.

        Returns
        -------
        _ValueT | _DefaultT
            The found entry or the default if any was provided.

        Raises
        ------
        CacheMissError
            If the entry wasn't found.

            This won't be raised if `default` is passed.
        EntryNotFound
            If the entry wasn't found and the the entry definitely doesn't exist.

            This won't be raised if `default` is passed.

            This is a specialisation of `CacheMissError` and thus may be
            caught as `CacheMissError and otherwise would need to be before
            `CacheMissError` in a try, multiple catch statement.
        """

    @abc.abstractmethod
    def iter_for_guild(self, guild_id: hikari.Snowflakeish, /) -> CacheIterator[_ValueT]:
        """Asynchronously iterate over the entries entries cached for a guild.

        !!! note
            For more information on how this is used, see the documentation for
            [hikari.iterators.LazyIterator][].

        Parameters
        ----------
        guild_id
            ID of the guild to iterate over the entries cached for.

        Returns
        -------
        CacheIterator[_ValueT]
            An asynchronous iterator of the entries cached for the specified guild.
        """

    @abc.abstractmethod
    def iter_all(self) -> CacheIterator[_ValueT]:
        """Asynchronously iterate over the globally cached entries for this resource.

        !!! note
            For more information on how this is used, see the documentation for
            [hikari.iterators.LazyIterator][].

        Returns
        -------
        CacheIterator[_ValueT]
            An asynchronous iterator of the entries cached globally for this resource.
        """


SfGuildBound = GuildBoundCache[hikari.Snowflakeish, _ValueT]
"""Alias of [tanjun.dependencies.GuildBoundCache][] where the key is a snowflake."""



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Command cooldown and concurrency limiters."""
from __future__ import annotations

__all__: list[str] = [
    "AbstractConcurrencyLimiter",
    "AbstractCooldownManager",
    "BucketResource",
    "ConcurrencyPostExecution",
    "ConcurrencyPreExecution",
    "CooldownPreExecution",
    "InMemoryConcurrencyLimiter",
    "InMemoryCooldownManager",
    "with_concurrency_limit",
    "with_cooldown",
]

import abc
import asyncio
import datetime
import enum
import logging
import typing

import alluka
import hikari

from .. import _internal
from .. import abc as tanjun
from .. import conversion
from .. import errors
from .. import hooks
from .._internal import localisation
from . import async_cache
from . import locales
from . import owners

if typing.TYPE_CHECKING:
    from collections import abc as collections

    from typing_extensions import Self

    _CommandT = typing.TypeVar("_CommandT", bound=tanjun.ExecutableCommand[typing.Any])
    _OtherCommandT = typing.TypeVar("_OtherCommandT", bound=tanjun.ExecutableCommand[typing.Any])
    _InnerResourceSig = collections.Callable[[], "_InnerResourceT"]


_InnerResourceT = typing.TypeVar("_InnerResourceT", bound="_InnerResourceProto")

_LOGGER: typing.Final[logging.Logger] = logging.getLogger("hikari.tanjun")


class AbstractCooldownManager(abc.ABC):
    """Interface used for managing command calldowns."""

    __slots__ = ()

    @abc.abstractmethod
    async def check_cooldown(
        self, bucket_id: str, ctx: tanjun.Context, /, *, increment: bool = False
    ) -> typing.Optional[datetime.datetime]:
        """Check if a bucket is on cooldown for the provided context.

        Parameters
        ----------
        bucket_id
            The cooldown bucket to check.
        ctx
            The context of the command.
        increment
            Whether this call should increment the bucket's use counter if
            it isn't depleted.

        Returns
        -------
        datetime.datetime | None
            When this command will next be usable for the provided context
            if it's in cooldown else [None][].
        """

    @abc.abstractmethod
    async def increment_cooldown(self, bucket_id: str, ctx: tanjun.Context, /) -> None:
        """Increment the cooldown of a cooldown bucket.

        Parameters
        ----------
        bucket_id
            The cooldown bucket's ID.
        ctx
            The context of the command.
        """


class AbstractConcurrencyLimiter(abc.ABC):
    """Interface used for limiting command concurrent usage."""

    __slots__ = ()

    @abc.abstractmethod
    async def try_acquire(self, bucket_id: str, ctx: tanjun.Context, /) -> bool:
        """Try to acquire a concurrency lock on a bucket.

        Parameters
        ----------
        bucket_id
            The concurrency bucket to acquire.
        ctx
            The context to acquire this resource lock with.

        Returns
        -------
        bool
            Whether the lock was acquired.
        """

    @abc.abstractmethod
    async def release(self, bucket_id: str, ctx: tanjun.Context, /) -> None:
        """Release a concurrency lock on a bucket."""


class BucketResource(int, enum.Enum):
    """Resource target types used within command calldowns and concurrency limiters."""

    USER = 0
    """A per-user resource bucket."""

    MEMBER = 1
    """A per-guild member resource bucket.

    When executed in a DM this will be per-DM.
    """

    CHANNEL = 2
    """A per-channel resource bucket."""

    PARENT_CHANNEL = 3
    """A per-parent channel resource bucket.

    For DM channels this will be per-DM, for guild channels with no parents
    this'll be per-guild.
    """

    # TODO: CATEGORY = 4

    TOP_ROLE = 5
    """A per-highest role resource bucket.

    When executed in a DM this will be per-DM, with this defaulting to
    targeting the @everyone role if they have no real roles.
    """

    GUILD = 6
    """A per-guild resource bucket.

    When executed in a DM this will be per-DM.
    """

    GLOBAL = 7
    """A global resource bucket."""


async def _try_get_role(
    cache: async_cache.SfCache[hikari.Role], role_id: hikari.Snowflake, /
) -> typing.Optional[hikari.Role]:
    try:
        return await cache.get(role_id)
    except async_cache.EntryNotFound:
        return None  # MyPy compat


async def _get_ctx_target(ctx: tanjun.Context, type_: BucketResource, /) -> hikari.Snowflake:
    if type_ is BucketResource.USER:
        return ctx.author.id

    if type_ is BucketResource.CHANNEL:
        return ctx.channel_id

    if type_ is BucketResource.PARENT_CHANNEL:
        channel: typing.Optional[hikari.PartialChannel]  # MyPy compat
        if ctx.guild_id is None:
            return ctx.channel_id

        if cached_channel := ctx.get_channel():
            return cached_channel.parent_id or ctx.guild_id

        channel_cache = ctx.get_type_dependency(async_cache.SfCache[hikari.PermissibleGuildChannel])
        if channel_cache and (channel := await channel_cache.get(ctx.channel_id, default=None)):
            return channel.parent_id or ctx.guild_id

        thread_cache = ctx.get_type_dependency(async_cache.SfCache[hikari.GuildThreadChannel])
        if thread_cache and (channel := await thread_cache.get(ctx.channel_id, default=None)):
            return channel.parent_id

        channel = await ctx.fetch_channel()
        assert isinstance(channel, hikari.GuildChannel)
        return channel.parent_id or ctx.guild_id

    if type_ is BucketResource.TOP_ROLE:
        if not ctx.guild_id:
            return ctx.channel_id

        # If they don't have a member object but this is in a guild context then we'll have to assume they
        # only have @everyone since they might be a webhook or something.
        if not ctx.member or len(ctx.member.role_ids) <= 1:  # If they only have 1 role ID then this is @everyone.
            return ctx.guild_id

        roles: collections.Iterable[hikari.Role] = ctx.member.get_roles()
        try_rest = not roles
        if try_rest and (role_cache := ctx.get_type_dependency(async_cache.SfCache[hikari.Role])):
            try:
                roles = filter(None, [await _try_get_role(role_cache, role_id) for role_id in ctx.member.role_ids])
                try_rest = False

            except async_cache.CacheMissError:
                pass

        if try_rest:
            roles = await ctx.member.fetch_roles()

        return next(iter(sorted(roles, key=lambda r: r.position, reverse=True))).id

    if type_ is BucketResource.GUILD:
        return ctx.guild_id or ctx.channel_id

    raise ValueError(f"Unexpected type {type_!s}")


def _now() -> datetime.datetime:
    return datetime.datetime.now(tz=datetime.timezone.utc)


class _Cooldown:
    __slots__ = ("counter", "limit", "reset_after", "resets_at")

    def __init__(self, *, limit: int, reset_after: datetime.timedelta) -> None:
        self.counter = 0
        self.limit = limit
        self.reset_after = reset_after
        self.resets_at = _now() + reset_after

    def has_expired(self) -> bool:
        # Expiration doesn't actually matter for cases where the limit is -1.
        return _now() >= self.resets_at

    def increment(self) -> Self:
        # A limit of -1 is special cased to mean no limit, so there's no need to increment the counter.
        if self.limit == -1:
            return self

        if self.counter == 0:
            self.resets_at = _now() + self.reset_after

        elif (current_time := _now()) >= self.resets_at:
            self.counter = 0
            self.resets_at = current_time + self.reset_after

        if self.counter < self.limit:
            self.counter += 1

        return self

    def must_wait_until(self) -> typing.Optional[datetime.datetime]:
        # A limit of -1 is special cased to mean no limit, so we don't need to wait.
        if self.limit == -1:
            return None

        if self.counter >= self.limit and self.resets_at > _now():
            return self.resets_at

        return None  # MyPy compat


class _MakeCooldown:
    __slots__ = ("limit", "reset_after")

    def __init__(self, *, limit: int, reset_after: datetime.timedelta) -> None:
        self.limit = limit
        self.reset_after = reset_after

    def __call__(self) -> _Cooldown:
        return _Cooldown(limit=self.limit, reset_after=self.reset_after)


class _InnerResourceProto(typing.Protocol):
    def has_expired(self) -> bool:
        raise NotImplementedError


class _BaseResource(abc.ABC, typing.Generic[_InnerResourceT]):
    __slots__ = ("make_resource",)

    def __init__(self, make_resource: _InnerResourceSig[_InnerResourceT], /) -> None:
        self.make_resource = make_resource

    @abc.abstractmethod
    def cleanup(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def copy(self) -> _BaseResource[_InnerResourceT]:
        raise NotImplementedError

    @abc.abstractmethod
    async def into_inner(self, ctx: tanjun.Context, /) -> _InnerResourceT:
        raise NotImplementedError

    @abc.abstractmethod
    async def try_into_inner(self, ctx: tanjun.Context, /) -> typing.Optional[_InnerResourceT]:
        raise NotImplementedError


class _FlatResource(_BaseResource[_InnerResourceT]):
    __slots__ = ("mapping", "resource")

    def __init__(self, resource: BucketResource, make_resource: _InnerResourceSig[_InnerResourceT], /) -> None:
        super().__init__(make_resource)
        self.mapping: dict[hikari.Snowflake, _InnerResourceT] = {}
        self.resource = resource

    async def try_into_inner(self, ctx: tanjun.Context, /) -> typing.Optional[_InnerResourceT]:
        return self.mapping.get(await _get_ctx_target(ctx, self.resource))

    async def into_inner(self, ctx: tanjun.Context, /) -> _InnerResourceT:
        target = await _get_ctx_target(ctx, self.resource)
        if resource := self.mapping.get(target):
            return resource

        resource = self.mapping[target] = self.make_resource()
        return resource

    def cleanup(self) -> None:
        for target_id, resource in self.mapping.copy().items():
            if resource.has_expired():
                del self.mapping[target_id]

    def copy(self) -> _FlatResource[_InnerResourceT]:
        return _FlatResource(self.resource, self.make_resource)


class _MemberResource(_BaseResource[_InnerResourceT]):
    __slots__ = ("dm_fallback", "mapping")

    def __init__(self, make_resource: _InnerResourceSig[_InnerResourceT], /) -> None:
        super().__init__(make_resource)
        self.dm_fallback: dict[hikari.Snowflake, _InnerResourceT] = {}
        self.mapping: dict[hikari.Snowflake, dict[hikari.Snowflake, _InnerResourceT]] = {}

    async def into_inner(self, ctx: tanjun.Context, /) -> _InnerResourceT:
        if not ctx.guild_id:
            if resource := self.dm_fallback.get(ctx.channel_id):
                return resource

            resource = self.dm_fallback[ctx.channel_id] = self.make_resource()
            return resource

        if (guild_mapping := self.mapping.get(ctx.guild_id)) is not None:
            if resource := guild_mapping.get(ctx.author.id):
                return resource

            resource = guild_mapping[ctx.author.id] = self.make_resource()
            return resource

        resource = self.make_resource()
        self.mapping[ctx.guild_id] = {ctx.author.id: resource}
        return resource

    async def try_into_inner(self, ctx: tanjun.Context, /) -> typing.Optional[_InnerResourceT]:
        if not ctx.guild_id:
            return self.dm_fallback.get(ctx.channel_id)

        if guild_mapping := self.mapping.get(ctx.guild_id):
            return guild_mapping.get(ctx.author.id)

        return None  # MyPy compat

    def cleanup(self) -> None:
        for guild_id, mapping in self.mapping.copy().items():
            for bucket_id, resource in mapping.copy().items():
                if resource.has_expired():
                    del mapping[bucket_id]

            if not mapping:
                del self.mapping[guild_id]

        for bucket_id, resource in self.dm_fallback.copy().items():
            if resource.has_expired():
                del self.dm_fallback[bucket_id]

    def copy(self) -> _MemberResource[_InnerResourceT]:
        return _MemberResource(self.make_resource)


class _GlobalResource(_BaseResource[_InnerResourceT]):
    __slots__ = ("bucket",)

    def __init__(self, make_resource: _InnerResourceSig[_InnerResourceT], /) -> None:
        super().__init__(make_resource)
        self.bucket = make_resource()

    async def try_into_inner(self, _: tanjun.Context, /) -> typing.Optional[_InnerResourceT]:
        return self.bucket

    async def into_inner(self, _: tanjun.Context, /) -> _InnerResourceT:
        return self.bucket

    def cleanup(self) -> None:
        pass

    def copy(self) -> _GlobalResource[_InnerResourceT]:
        return _GlobalResource(self.make_resource)


def _to_bucket(
    resource: BucketResource, make_resource: _InnerResourceSig[_InnerResourceT], /
) -> _BaseResource[_InnerResourceT]:
    if resource is BucketResource.MEMBER:
        return _MemberResource(make_resource)

    if resource is BucketResource.GLOBAL:
        return _GlobalResource(make_resource)

    return _FlatResource(resource, make_resource)


class InMemoryCooldownManager(AbstractCooldownManager):
    """In-memory standard implementation of [AbstractCooldownManager][tanjun.dependencies.AbstractCooldownManager].

    Examples
    --------
    [InMemoryCooldownManager.set_bucket][tanjun.dependencies.InMemoryCooldownManager.set_bucket]
    may be used to set the cooldown for a specific bucket:

    ```py
    (
        InMemoryCooldownManager()
        # Set the default bucket template to a per-user 10 uses per-60 seconds cooldown.
        .set_bucket("default", tanjun.BucketResource.USER, 10, 60)
        # Set the "moderation" bucket to a per-guild 100 uses per-5 minutes cooldown.
        .set_bucket("moderation", tanjun.BucketResource.GUILD, 100, datetime.timedelta(minutes=5))
        .set_bucket()
        # add_to_client will setup the cooldown manager (setting it as an
        # injected dependency and registering callbacks to manage it).
        .add_to_client(client)
    )
    ```
    """

    __slots__ = ("_buckets", "_default_bucket_template", "_gc_task")

    def __init__(self) -> None:
        self._buckets: dict[str, _BaseResource[_Cooldown]] = {}
        self._default_bucket_template: _BaseResource[_Cooldown] = _FlatResource(
            BucketResource.USER, _MakeCooldown(limit=2, reset_after=datetime.timedelta(seconds=5))
        )
        self._gc_task: typing.Optional[asyncio.Task[None]] = None

    def _get_or_default(self, bucket_id: str, /) -> _BaseResource[_Cooldown]:
        if bucket := self._buckets.get(bucket_id):
            return bucket

        _LOGGER.info("No cooldown found for %r, falling back to 'default' bucket", bucket_id)
        bucket = self._buckets[bucket_id] = self._default_bucket_template.copy()
        return bucket

    async def _gc(self) -> None:
        while True:
            await asyncio.sleep(10)
            for bucket in self._buckets.values():
                bucket.cleanup()

    def add_to_client(self, client: tanjun.Client, /) -> None:
        """Add this cooldown manager to a tanjun client.

        !!! note
            This registers the manager as a type dependency and manages opening
            and closing the manager based on the client's life cycle.

        Parameters
        ----------
        client
            The client to add this cooldown manager to.
        """
        client.set_type_dependency(AbstractCooldownManager, self)
        client.add_client_callback(tanjun.ClientCallbackNames.STARTING, self.open)
        client.add_client_callback(tanjun.ClientCallbackNames.CLOSING, self.close)
        if client.is_alive:
            assert client.loop is not None
            self.open(_loop=client.loop)

    async def check_cooldown(
        self, bucket_id: str, ctx: tanjun.Context, /, *, increment: bool = False
    ) -> typing.Optional[datetime.datetime]:
        # <<inherited docstring from AbstractCooldownManager>>.
        bucket: typing.Optional[_Cooldown]
        if increment:
            bucket = await self._get_or_default(bucket_id).into_inner(ctx)
            if cooldown := bucket.must_wait_until():
                return cooldown

            bucket.increment()
            return None

        if (resource := self._buckets.get(bucket_id)) and (bucket := await resource.try_into_inner(ctx)):
            return bucket.must_wait_until()

        return None  # MyPy compat

    async def increment_cooldown(self, bucket_id: str, ctx: tanjun.Context, /) -> None:
        # <<inherited docstring from AbstractCooldownManager>>.
        (await self._get_or_default(bucket_id).into_inner(ctx)).increment()

    def close(self) -> None:
        """Stop the cooldown manager.

        Raises
        ------
        RuntimeError
            If the cooldown manager is not running.
        """
        if not self._gc_task:
            raise RuntimeError("Cooldown manager is not active")

        self._gc_task.cancel()
        self._gc_task = None

    def open(self, *, _loop: typing.Optional[asyncio.AbstractEventLoop] = None) -> None:
        """Start the cooldown manager.

        Raises
        ------
        RuntimeError
            If the cooldown manager is already running.
            If called in a thread with no running event loop.
        """
        if self._gc_task:
            raise RuntimeError("Cooldown manager is already running")

        self._gc_task = (_loop or asyncio.get_running_loop()).create_task(self._gc())

    def disable_bucket(self, bucket_id: str, /) -> Self:
        """Disable a cooldown bucket.

        This will stop the bucket from ever hitting a cooldown and also
        prevents the bucket from defaulting.

        !!! note
            "default" is a special `bucket_id` which is used as a template for
            unknown bucket IDs.

        Parameters
        ----------
        bucket_id
            The bucket to disable.

        Returns
        -------
        Self
            This cooldown manager to allow for chaining.
        """
        # A limit of -1 is special cased to mean no limit and reset_after is ignored in this scenario.
        bucket = self._buckets[bucket_id] = _GlobalResource(_MakeCooldown(limit=-1, reset_after=datetime.timedelta(-1)))
        if bucket_id == "default":
            self._default_bucket_template = bucket.copy()

        return self

    def set_bucket(
        self,
        bucket_id: str,
        resource: BucketResource,
        limit: int,
        reset_after: typing.Union[int, float, datetime.timedelta],
        /,
    ) -> Self:
        """Set the cooldown for a specific bucket.

        !!! note
            "default" is a special `bucket_id` which is used as a template for
            unknown bucket IDs.

        Parameters
        ----------
        bucket_id
            The ID of the bucket to set the cooldown for.
        resource
            The type of resource to target for the cooldown.
        limit
            The number of uses per cooldown period.
        reset_after
            The cooldown period.

        Returns
        -------
        Self
            The cooldown manager to allow call chaining.

        Raises
        ------
        ValueError
            If an invalid resource type is given.
            If reset_after or limit are negative, 0 or invalid.
            if limit is less 0 or negative.
        """
        if not isinstance(reset_after, datetime.timedelta):
            reset_after = datetime.timedelta(seconds=reset_after)

        if reset_after <= datetime.timedelta():
            raise ValueError("reset_after must be greater than 0 seconds")

        if limit <= 0:
            raise ValueError("limit must be greater than 0")

        bucket = self._buckets[bucket_id] = _to_bucket(
            BucketResource(resource), _MakeCooldown(limit=limit, reset_after=reset_after)
        )
        if bucket_id == "default":
            self._default_bucket_template = bucket.copy()

        return self


class CooldownPreExecution:
    """Pre-execution hook used to manage a command's cooldowns.

    To avoid race-conditions this handles both erroring when the bucket is hit
    instead and incrementing the bucket's use counter.
    """

    __slots__ = ("_bucket_id", "_error", "_error_message", "_owners_exempt", "__weakref__")

    def __init__(
        self,
        bucket_id: str,
        /,
        *,
        error: typing.Optional[collections.Callable[[str, datetime.datetime], Exception]] = None,
        error_message: typing.Union[
            str, collections.Mapping[str, str]
        ] = "This command is currently in cooldown. Try again {cooldown}.",
        owners_exempt: bool = True,
    ) -> None:
        """Initialise a pre-execution cooldown command hook.

        Parameters
        ----------
        bucket_id
            The cooldown bucket's ID.
        error
            Callback used to create a custom error to raise if the check fails.

            This should two arguments one of type [str][] and [datetime.datetime][]
            where the first is the limiting bucket's ID and the second is when said
            bucket can be used again.

            This takes priority over `error_message`.
        error_message
            The error message to send in response as a command error if the check fails.

            This supports [localisation][] and uses the check name
            `"tanjun.cooldown"` for global overrides.
        owners_exempt
            Whether owners should be exempt from the cooldown.
        """
        self._bucket_id = bucket_id
        self._error = error
        self._error_message = localisation.MaybeLocalised("error_message", error_message)
        self._owners_exempt = owners_exempt

    async def __call__(
        self,
        ctx: tanjun.Context,
        /,
        cooldowns: alluka.Injected[AbstractCooldownManager],
        *,
        localiser: typing.Optional[locales.AbstractLocaliser] = None,
        owner_check: alluka.Injected[typing.Optional[owners.AbstractOwners]],
    ) -> None:
        if self._owners_exempt:
            if not owner_check:
                _LOGGER.info("No `AbstractOwners` dependency found, disabling owner exemption for cooldown check")
                self._owners_exempt = False

            elif await owner_check.check_ownership(ctx.client, ctx.author):
                return

        if wait_until := await cooldowns.check_cooldown(self._bucket_id, ctx, increment=True):
            if self._error:
                raise self._error(self._bucket_id, wait_until) from None

            wait_until_repr = conversion.from_datetime(wait_until, style="R")
            message = self._error_message.localise(ctx, localiser, "check", "tanjun.cooldown", cooldown=wait_until_repr)
            raise errors.CommandError(message)


def with_cooldown(
    bucket_id: str,
    /,
    *,
    error: typing.Optional[collections.Callable[[str, datetime.datetime], Exception]] = None,
    error_message: typing.Union[
        str, collections.Mapping[str, str]
    ] = "This command is currently in cooldown. Try again {cooldown}.",
    follow_wrapped: bool = False,
    owners_exempt: bool = True,
) -> collections.Callable[[_CommandT], _CommandT]:
    """Add a pre-execution hook used to manage a command's cooldown through a decorator call.

    !!! warning
        Cooldowns will only work if there's a setup injected
        [AbstractCooldownManager][tanjun.dependencies.InMemoryCooldownManager] dependency with
        [InMemoryCooldownManager][tanjun.dependencies.InMemoryCooldownManager]
        being usable as a standard in-memory cooldown manager.

    Parameters
    ----------
    bucket_id
        The cooldown bucket's ID.
    error
        Callback used to create a custom error to raise if the check fails.

        This should two arguments one of type [str][] and [datetime.datetime][]
        where the first is the limiting bucket's ID and the second is when said
        bucket can be used again.

        This takes priority over `error_message`.
    error_message
        The error message to send in response as a command error if the check fails.

        This supports [localisation][] and uses the check name
        `"tanjun.cooldown"` for global overrides.
    follow_wrapped
        Whether to also add this check to any other command objects this
        command wraps in a decorator call chain.
    owners_exempt
        Whether owners should be exempt from the cooldown.

    Returns
    -------
    collections.abc.Callable[[tanjun.abc.ExecutableCommand], tanjun.abc.ExecutableCommand]
        A decorator that adds a [CooldownPreExecution][tanjun.dependencies.CooldownPreExecution]
        hook to the command.
    """
    pre_execution = CooldownPreExecution(
        bucket_id, error=error, error_message=error_message, owners_exempt=owners_exempt
    )

    def decorator(command: _OtherCommandT, /, *, _recursing: bool = False) -> _OtherCommandT:
        hooks_ = command.hooks
        if not hooks_:
            hooks_ = hooks.AnyHooks()
            command.set_hooks(hooks_)

        hooks_.add_pre_execution(pre_execution)
        if follow_wrapped and not _recursing:
            for wrapped in _internal.collect_wrapped(command):
                decorator(wrapped, _recursing=True)

        return command

    return decorator


class _ConcurrencyLimit:
    __slots__ = ("counter", "limit")

    def __init__(self, limit: int, /) -> None:
        self.counter = 0
        self.limit = limit

    def acquire(self) -> bool:
        if self.counter < self.limit:
            self.counter += 1
            return True

        # A limit of -1 means unlimited so we don't need to keep count.
        if self.limit == -1:
            return True

        return False

    def release(self) -> None:
        if self.counter > 0:
            self.counter -= 1
            return

        # A limit of -1 means unlimited so we don't need to keep count.
        if self.limit == -1:
            return

        raise RuntimeError("Cannot release a limit that has not been acquired, this should never happen")

    def has_expired(self) -> bool:
        # Expiration doesn't actually matter for cases where the limit is -1.
        return self.counter == 0


class InMemoryConcurrencyLimiter(AbstractConcurrencyLimiter):
    """In-memory standard implementation of [AbstractConcurrencyLimiter][tanjun.dependencies.AbstractConcurrencyLimiter].

    Examples
    --------
    [InMemoryConcurrencyLimiter.set_bucket][tanjun.dependencies.InMemoryConcurrencyLimiter.set_bucket]
    may be used to set the concurrency limits for a specific bucket:

    ```py
    (
        InMemoryConcurrencyLimiter()
        # Set the default bucket template to 10 concurrent uses of the command per-user.
        .set_bucket("default", tanjun.BucketResource.USER, 10)
        # Set the "moderation" bucket with a limit of 5 concurrent uses per-guild.
        .set_bucket("moderation", tanjun.BucketResource.GUILD, 5)
        .set_bucket()
        # add_to_client will setup the concurrency manager (setting it as an
        # injected dependency and registering callbacks to manage it).
        .add_to_client(client)
    )
    ```
    """

    __slots__ = ("_acquiring_ctxs", "_buckets", "_default_bucket_template", "_gc_task")

    def __init__(self) -> None:
        self._acquiring_ctxs: dict[tuple[str, tanjun.Context], _ConcurrencyLimit] = {}
        self._buckets: dict[str, _BaseResource[_ConcurrencyLimit]] = {}
        self._default_bucket_template: _BaseResource[_ConcurrencyLimit] = _FlatResource(
            BucketResource.USER, lambda: _ConcurrencyLimit(1)
        )
        self._gc_task: typing.Optional[asyncio.Task[None]] = None

    async def _gc(self) -> None:
        while True:
            await asyncio.sleep(10)
            for bucket in self._buckets.values():
                bucket.cleanup()

    def add_to_client(self, client: tanjun.Client, /) -> None:
        """Add this concurrency manager to a tanjun client.

        !!! note
            This registers the manager as a type dependency and manages opening
            and closing the manager based on the client's life cycle.

        Parameters
        ----------
        client
            The client to add this concurrency manager to.
        """
        client.set_type_dependency(AbstractConcurrencyLimiter, self)
        client.add_client_callback(tanjun.ClientCallbackNames.STARTING, self.open)
        client.add_client_callback(tanjun.ClientCallbackNames.CLOSING, self.close)
        if client.is_alive:
            assert client.loop is not None
            self.open(_loop=client.loop)

    def close(self) -> None:
        """Stop the concurrency manager.

        Raises
        ------
        RuntimeError
            If the concurrency manager is not running.
        """
        if not self._gc_task:
            raise RuntimeError("Concurrency manager is not active")

        self._gc_task.cancel()
        self._gc_task = None

    def open(self, *, _loop: typing.Optional[asyncio.AbstractEventLoop] = None) -> None:
        """Start the concurrency manager.

        Raises
        ------
        RuntimeError
            If the concurrency manager is already running.
            If called in a thread with no running event loop.
        """
        if self._gc_task:
            raise RuntimeError("Concurrency manager is already running")

        self._gc_task = (_loop or asyncio.get_running_loop()).create_task(self._gc())

    async def try_acquire(self, bucket_id: str, ctx: tanjun.Context, /) -> bool:
        # <<inherited docstring from AbstractConcurrencyLimiter>>.
        bucket = self._buckets.get(bucket_id)
        if not bucket:
            _LOGGER.info("No concurrency limit found for %r, falling back to 'default' bucket", bucket_id)
            bucket = self._buckets[bucket_id] = self._default_bucket_template.copy()

        # incrementing a bucket multiple times for the same context could lead
        # to weird edge cases based on how we internally track this, so we
        # internally de-duplicate this.
        elif (bucket_id, ctx) in self._acquiring_ctxs:
            return True  # This won't ever be the case if it just had to make a new bucket, hence the elif.

        if result := (limit := await bucket.into_inner(ctx)).acquire():
            self._acquiring_ctxs[(bucket_id, ctx)] = limit

        return result

    async def release(self, bucket_id: str, ctx: tanjun.Context, /) -> None:
        # <<inherited docstring from AbstractConcurrencyLimiter>>.
        if limit := self._acquiring_ctxs.pop((bucket_id, ctx), None):
            limit.release()

    def disable_bucket(self, bucket_id: str, /) -> Self:
        """Disable a concurrency limit bucket.

        This will stop the bucket from ever hitting a concurrency limit
        and also prevents the bucket from defaulting.

        !!! note
            "default" is a special `bucket_id` which is used as a template for
            unknown bucket IDs.

        Parameters
        ----------
        bucket_id
            The bucket to disable.

        Returns
        -------
        Self
            This concurrency manager to allow for chaining.
        """
        bucket = self._buckets[bucket_id] = _GlobalResource(lambda: _ConcurrencyLimit(-1))
        if bucket_id == "default":
            self._default_bucket_template = bucket.copy()

        return self

    def set_bucket(self, bucket_id: str, resource: BucketResource, limit: int, /) -> Self:
        """Set the concurrency limit for a specific bucket.

        !!! note
            "default" is a special `bucket_id` which is used as a template for
            unknown bucket IDs.

        Parameters
        ----------
        bucket_id
            The ID of the bucket to set the concurrency limit for.
        resource
            The type of resource to target for the concurrency limit.
        limit
            The maximum number of concurrent uses to allow.

        Returns
        -------
        Self
            The concurrency manager to allow call chaining.

        Raises
        ------
        ValueError
            If an invalid resource type is given.
            if limit is less 0 or negative.
        """
        if limit <= 0:
            raise ValueError("limit must be greater than 0")

        bucket = self._buckets[bucket_id] = _to_bucket(BucketResource(resource), lambda: _ConcurrencyLimit(limit))
        if bucket_id == "default":
            self._default_bucket_template = bucket.copy()

        return self


class ConcurrencyPreExecution:
    """Pre-execution hook used to acquire a bucket concurrency limiter.

    !!! note
        For a concurrency limiter to work properly, both
        [ConcurrencyPreExecution][tanjun.dependencies.ConcurrencyPreExecution]
        and [ConcurrencyPostExecution][tanjun.dependencies.ConcurrencyPostExecution]
        hooks must be registered for a command scope.
    """

    __slots__ = ("_bucket_id", "_error", "_error_message", "__weakref__")

    def __init__(
        self,
        bucket_id: str,
        /,
        *,
        error: typing.Optional[collections.Callable[[str], Exception]] = None,
        error_message: typing.Union[
            str, collections.Mapping[str, str]
        ] = "This resource is currently busy; please try again later.",
    ) -> None:
        """Initialise a concurrency pre-execution hook.

        Parameters
        ----------
        bucket_id
            The concurrency limit bucket's ID.
        error
            Callback used to create a custom error to raise if the check fails.

            This should two one [str][] argument which is the limiting bucket's ID.

            This takes priority over `error_message`.
        error_message
            The error message to send in response as a command error if this fails
            to acquire the concurrency limit.

            This supports [localisation][] and uses the check name
            `"tanjun.concurrency"` for global overrides.
        """
        self._bucket_id = bucket_id
        self._error = error
        self._error_message = localisation.MaybeLocalised("error_message", error_message)

    async def __call__(
        self,
        ctx: tanjun.Context,
        /,
        limiter: alluka.Injected[AbstractConcurrencyLimiter],
        *,
        localiser: typing.Optional[locales.AbstractLocaliser] = None,
    ) -> None:
        if not await limiter.try_acquire(self._bucket_id, ctx):
            if self._error:
                raise self._error(self._bucket_id) from None

            message = self._error_message.localise(ctx, localiser, "check", "tanjun.concurrency")
            raise errors.CommandError(message) from None


class ConcurrencyPostExecution:
    """Post-execution hook used to release a bucket concurrency limiter.

    !!! note
        For a concurrency limiter to work properly, both
        [ConcurrencyPreExecution][tanjun.dependencies.ConcurrencyPreExecution]
        and [ConcurrencyPostExecution][tanjun.dependencies.ConcurrencyPostExecution]
        hooks must be registered for a command scope.
    """

    __slots__ = ("_bucket_id", "__weakref__")

    def __init__(self, bucket_id: str, /) -> None:
        """Initialise a concurrency post-execution hook.

        Parameters
        ----------
        bucket_id
            The concurrency limit bucket's ID.
        """
        self._bucket_id = bucket_id

    async def __call__(self, ctx: tanjun.Context, /, limiter: alluka.Injected[AbstractConcurrencyLimiter]) -> None:
        await limiter.release(self._bucket_id, ctx)


def with_concurrency_limit(
    bucket_id: str,
    /,
    *,
    error: typing.Optional[collections.Callable[[str], Exception]] = None,
    error_message: typing.Union[
        str, collections.Mapping[str, str]
    ] = "This resource is currently busy; please try again later.",
    follow_wrapped: bool = False,
) -> collections.Callable[[_CommandT], _CommandT]:
    """Add the hooks used to manage a command's concurrency limit through a decorator call.

    !!! warning
        Concurrency limiters will only work if there's a setup injected
        [AbstractConcurrencyLimiter][tanjun.dependencies.AbstractConcurrencyLimiter] dependency with
        [InMemoryConcurrencyLimiter][tanjun.dependencies.InMemoryConcurrencyLimiter]
        being usable as a standard in-memory concurrency manager.

    Parameters
    ----------
    bucket_id
        The concurrency limit bucket's ID.
    error
        Callback used to create a custom error to raise if the check fails.

        This should two one [str][] argument which is the limiting bucket's ID.

        This takes priority over `error_message`.
    error_message
        The error message to send in response as a command error if this fails
        to acquire the concurrency limit.

        This supports [localisation][] and uses the check name
        `"tanjun.concurrency"` for global overrides.
    follow_wrapped
        Whether to also add this check to any other command objects this
        command wraps in a decorator call chain.

    Returns
    -------
    collections.abc.Callable[[tanjun.abc.ExecutableCommand], tanjun.abc.ExecutableCommand]
        A decorator that adds the concurrency limiter hooks to a command.
    """
    pre_execution = ConcurrencyPreExecution(bucket_id, error=error, error_message=error_message)
    post_execution = ConcurrencyPostExecution(bucket_id)

    def decorator(command: _OtherCommandT, /, *, _recursing: bool = False) -> _OtherCommandT:
        hooks_ = command.hooks
        if not hooks_:
            hooks_ = hooks.AnyHooks()
            command.set_hooks(hooks_)

        hooks_.add_pre_execution(pre_execution).add_post_execution(post_execution)
        if follow_wrapped and not _recursing:
            for wrapped in _internal.collect_wrapped(command):
                decorator(wrapped, _recursing=True)

        return command

    return decorator
